import unittest
from bffacilities.socketserver import TcpSocketServer, SocketClient, logger
from bffacilities.utils import changeLoggerLevel
import logging
changeLoggerLevel(logger, logging.DEBUG)
import time
import threading
import asyncio
import urllib.request

async def createClient(id, address):
    # client = SocketClient.createTcpClient(address)
    keyword = "python"

    url = f"https://{address[0]}:{address[1]}/"
    try:
        i = 0
        while i < 50:
            resp = urllib.request.urlopen(url, timeout=2)
            if resp.getcode() == 408:
                print(id, "-- timeout")
            else:
                data = resp.read().decode("utf-8")
                # print(id, "--", i, data)
            # client.sock.send(data.encode())
            await asyncio.sleep(0) 
            i += 1
    except (KeyboardInterrupt, SystemExit):
        pass
    except Exception as e:
        print("exception:", e)
    print("Client Done ", id)
async def main(args):
    tasks = []
    for i in range(args["count"]):
        tasks.append(createClient(i, (args["host"], args["port"])))
    print("***start**")
    try:
        await asyncio.gather(*tasks)
    except (KeyboardInterrupt, SystemExit):
        # asyncio.
        pass
    

if __name__ == "__main__":
    import argparse, sys
    parser = argparse.ArgumentParser("TestTcpSocketClient")
    parser.add_argument("--port", type=int, default=54134, help="set port")
    parser.add_argument("--host", type=str, default="127.0.0.1", help="set host")
    parser.add_argument("--count", type=int, default=10, help="parallel count")
    args = vars(parser.parse_args(sys.argv[1:]))
    
    loop = asyncio.new_event_loop()
    loop.run_until_complete(main(args))




import unittest
from bffacilities.socketserver import TcpSocketServer, SocketClient, logger
from bffacilities.utils import changeLoggerLevel
import logging
changeLoggerLevel(logger, logging.DEBUG)

class TestTcpSocketServer(unittest.TestCase):

    def __init__(self, methodName):
        super().__init__(methodName)

    def setUp(self):
        self.tcpServer = TcpSocketServer("0.0.0.0", 54134)
        super().setUp()

    def tearDown(self):
        self.tcpServer.stop()
        self.tcpServer.shutdown()
        return super().tearDown()

    def test_connect(self):
        self.tcpServer.serve_once()
        client = SocketClient.createTcpClient(("127.0.0.1", 54134))
        client.connect()

        self.tcpServer.serve_once()
        client.sock.send(b"[This is a test]")

        self.tcpServer.serve_once(0.05)
        # self.tcpServer.stop()
        client.shutdown()

if __name__ == "__main__":
    # unittest.main()
    tcpServer = TcpSocketServer("0.0.0.0", 54134)
    tcpServer.serve_forever()


import sys
from bleak import BleakScanner, BleakClient
import asyncio
from time import sleep
import threading, queue
import socket
from datetime import datetime
import json
from bffacilities.utils import createLogger
import logging
logger = createLogger("bt", level=logging.DEBUG,\
    savefile=False, stream=True, timeformat="%H:%M:%S" )

if sys.platform == "win32":
    from Windows.Devices.Bluetooth.Advertisement import BluetoothLEAdvertisementFilter, BluetoothLEAdvertisement 
else:
    raise ValueError("Unsupported")

class BluetoothWrapper(object):
    """封装蓝牙操作
    """
    IO_DATA_CHAR_UUID_W   = "0000ffe1-0000-1000-8000-00805f9b34fb"
    IO_DATA_CHAR_UUID_W2  = "0000ffe2-0000-1000-8000-00805f9b34fb"

    def __init__(self, **kwargs):
        self._btMsgProcessors = []
        processor = kwargs.get("btMsgProcessor")
        if processor is not None:
            self._btMsgProcessors.append(processor)
        self.loop = asyncio.get_event_loop()
        self._msgToBtQueue = queue.Queue()
        self._cmdQueue = queue.Queue()

        self._btClient = None
        self._btConnected = False
        self._devices = { } # key address
        self._lastAddress = None

    @property
    def is_connected(self):
        return self._btConnected
    @property
    def btMsgProcessor(self):
        return self._btMsgProcessors
    
    @btMsgProcessor.setter
    def btMsgProcessor(self, p):
        """
        Callback

            btMsgProcess(data: bytes)
        """
        self._btMsgProcessors.append(p)
    
    def export_deviceInfo(self):
        data = []
        for device in self._devices.values():
            name = device.name if hasattr(device, "name") else None
                
            data.append({
                "address": device.address,
                "name": name,
                "rssi": device.rssi
            })
        data = json.dumps(data)
        return data

    def syncDeviceToOther(self):
        """should be rewrite by subclass
        """
        pass

    def _onRecvBtMsg(self, sender: int, data: bytes):
        """
        Args
            data bytes
        """
        # logger.debug(f"Recv BT & Send: {len(data)}")
        for p in self._btMsgProcessors:
            p(data)

    def listDevice(self, timeout = 10):
        async def _list(timeout):
            devices = await BleakScanner.discover(timeout=timeout)
            for d in devices:
                if d.address not in self._devices:
                    self._devices[d.address] = d
            logger.info(f"{self._devices}")
            self.syncDeviceToOther()
        self.loop.run_until_complete(_list(timeout))
    
    def findDevice(self, name = None, seconds = 5.0):
        self.loop.run_until_complete(self.monitDevice(name, seconds))

    async def monitDevice(self, name, seconds=0):
        """
        Args:
            name  device name such as "BTDEVICE1"
        """
        bleFilter = BluetoothLEAdvertisementFilter()
        if name:
            ad = BluetoothLEAdvertisement()
            ad.put_LocalName(name)
            bleFilter.put_Advertisement(ad)

        scanner = BleakScanner(AdvertisementFilter=bleFilter)
        scanner.register_detection_callback(self._detection_callback)
        # await scanner.set_scanning_filter()
        await scanner.start()
        if seconds <= 5:
            while self._btConnected:
                await asyncio.sleep(10)
        else:
            await asyncio.sleep(seconds)
        await scanner.stop()
        logger.info("[BT] Monit Device Done")

    def _detection_callback(self, device, advertisement_data):
        if len(advertisement_data.local_name) == 0:
            return
        if device.address not in self._devices:
            logger.info(f"First [{datetime.now()}] RSSI: {device.rssi}, {advertisement_data.local_name}, {device.address}")
        device.name = advertisement_data.local_name
        self._devices[device.address] = device
        self.syncDeviceToOther()

    async def _connect(self, address):
        """after function exits, bluetooth device will be disconnected
        """
        if self._btClient is not None:
            await self._btClient.disconnect()

        self._btClient = BleakClient(address)
        logger.info(f"[BT] Connecting Bluetooth: {address}")
        # service = await self._btClient.get_services()
        # model_number = await self._btClient.read_gatt_char(BluetoothWrapper.IO_DATA_CHAR_UUID_W )
        # print(f"Model Number {model_number}: {''.join(map(chr, model_number))}")
        try:
            self._btConnected = await self._btClient.connect()
            logger.info(f"[BT] Client {self._btClient} connected: {self._btConnected}")
            if self._btConnected:
                await self._btClient.start_notify(\
                    BluetoothWrapper.IO_DATA_CHAR_UUID_W, self._onRecvBtMsg)
        except Exception as e:
            self._btConnected = False
            logger.warning(f"[BT] Connect failed: {e}")
        # finally:
            # await self._btClient.disconnect()


    async def _sendMsgToBt(self):
        if not self._msgToBtQueue.empty():
            sData = self._msgToBtQueue.get()
            try:
                await self._btClient.write_gatt_char(BluetoothWrapper.IO_DATA_CHAR_UUID_W, sData)
            except : 
                pass
            logger.debug(f"SendData Bt {sData}")
        else:
            await asyncio.sleep(0)
    
    def sendMsgToBt(self):
        self.loop.run_until_complete(self._sendMsgToBt())

    async def _connectAndListen(self, address):
        await self._connect(address)
        try:
            while self._btConnected:
                await self._sendMsgToBt()
        except Exception as e:
            self._btConnected = False
            logger.warning(f"[BT] Exception Listen: {e}")
        finally:
            await self._btClient.disconnect()
        logger.info("[BT] connect process done")

    def connectToBLE(self, address, listen=False):
        """
        Args: 
            address:  MAC address

            listen 
            if True, the event loop will run always and stop main thread

            if False, you have to call `sendMsgToBt` mannually if there is 
            pending msg needs to be sent to bluetooth

        Note:
            If a bluetooth is connected, it should be disconnect first
        """
        if self._btConnected:
            if address != self._lastAddress:
                self.disconnect()
            else:
                logger.warning("Bluetooth Device is already connected")
                return

        self._lastAddress = address
        if listen:
            self.loop.run_until_complete(self._connectAndListen(address))
        else:
            self.loop.run_until_complete(self._connect(address))

    def disconnect(self):
        if self._btConnected:
            logger.info("[BT] Stoping bluetooth, wait for a moment")
            async def _stop():
                await self._btClient.disconnect()
            self.loop.run_until_complete(_stop())
            logger.info("[BT] bluetooth stopped")
        else:
            logger.warning("[BT] Not Connected")


    def _processCmd(self, cmd):
        """
        {
            "action": "listDevice",
            "timeout": "60", // seconds
        }

        {
            "action": "findDevice",
            "name": "CMPS", // or no name for find all device
            "timeout": 30, // default 30
        }

        {
            "action": "disconnect" // disconnect if a device has connected
        }

        {
            "action": "quit" // quit program
        }
        """
        # logger.debug(f"Cmd { threading.get_ident()}")
        action = cmd["action"]
        if action == "connect":
            if "address" in cmd:
                self.connectToBLE(cmd["address"])
        elif action == "listDevice":
            self.listDevice(timeout= cmd.get("timeout", 10))
        elif action == "findDevice":
            self.findDevice(name=cmd.get("name"), seconds= cmd.get("timeout", 20))
        elif action == "disconnect":
            self.disconnect()
        elif action == "quit":
            self.disconnect()
            self._running = False
            logger.info("About to quit")
        else:
            logger.warning("Unsupported cmd", action)

    def processCmd(self):
        """注意，用于蓝牙接收和发送的操作，
        该方法应该在创建 BluetoothWrapper 的线程中调用
        """
        if not self._cmdQueue.empty():
            self._processCmd(self._cmdQueue.get()) 

from bffacilities.socketserver import UdpSocketServer

class BluetoothMgr(BluetoothWrapper) :
    """
    @See bffacilities.socketserver.UdpSocketServer

    默认绑定到 UDP 12001 地址上，用于接收其他模块的控制信息

    """

    def __init__(self, port = 12001, broadcast = True, **kwargs):
        super().__init__(**kwargs)
        self._bind = port
        self.sockServer = UdpSocketServer()
        self.sockServer.createListener(port = self._bind, ip="127.0.0.1")
        self.sockServer.addProcessor(self.recvCtlMsg)
        self._running = True
        if kwargs.get("btMsgProcessor") is None:
            self.btMsgProcessor = self.sockServer.send

    def transfer(self):
        self.btMsgProcessor = self.sockServer.send
    @property
    def is_running(self):
        return self._running

    def recvCtlMsg(self, msg: bytes, addr: tuple):
        # print(msg, addr)
        if msg.startswith(b"{"):
            try:
                cmd = json.loads(msg.decode())
                self._processCmd(cmd)
            except: pass
        else:
            self._msgToBtQueue.put(msg)

    def syncDeviceToOther(self):
        data = self.export_deviceInfo()
        self._sockSender.sendto(data.encode() + b"\n\n")

    def start(self):
        """start an loop in main thread 
        start another thread for udp server
        """
        try:
            while self._running:
                self.sendMsgToBt()
                self.sockServer.serve_once()
                sleep(0.1)
        except (KeyboardInterrupt, SystemExit):
            self.stop()
    
    start_one = start
    
    def stop(self):
        self.disconnect()
        self._running = False
        logger.info("==========Bt Mgr Quit===========")


if __name__ == "__main__":
    btMgr = BluetoothMgr()
    btMgr.start()


#!/usr/env/bin python3
# -*- coding: utf-8 -*-

'''
 # @ Author: brifuture
 # @ Create Time: 2021-04-02 15:33
 # @ Modified by: brifuture
 # @ Modified time: 2021-06-11 11:38
 # @ Description: This File is created and the program is coded by BriFuture. (c) All rights reserved.
    Use as script entry
 '''

import sys, os
import os.path as osp
from ._constants import BFF_ROOT_PATH as RootPath

from pathlib import Path
from .utils import initGetText
from ._globals import _logger as logger
from . import __version__
from ._plugin import load_scripts
import json

#  add custom sub-commands in this file
try:
    with open(osp.join(RootPath, "myscripts/meta.json")) as f:
        availableCmds = json.load(f)
except:
    availableCmds = {}
    logger.warning(f"RootDir ({RootPath}) dose not contain meta " )
    
availableCmds["tray"] = "tray"

tr = initGetText("bffacility")
def onError(parser):
    logger.info(tr('Available sub commands: '))
    cmds = ""
    for i, c in enumerate(availableCmds):
        cmds += f"{i+1}: {c}  \t"
    logger.info(f"{cmds} \n")
    parser.print_usage()

def loadPriScript(arg):
    """arg contain subscripts:
    for example: `bff pri test -h`
    arg should be `test -h`
    """
    if len(arg) < 1:
        logger.critical("Not Supported!")
        return
    pripath = Path(RootPath) / "../_pri"
    pripath = pripath.resolve()
    try:
        subscript = arg[0]
        p, ss = osp.split(subscript)
        pripath = Path(osp.join(pripath, p))
        if ss.endswith(".py"):
            ss = ss[:-3]
        logger.info(f"{ss}, {pripath}")
        arg = arg[1:]
        load_scripts(ss, arg, pripath)
    except Exception as e:
        logger.warning("Error Running Pri Scripts: ", e) 

def loadTorch(arg):
    if len(arg) < 1:
        logger.critical("Not Supported!")
        return
    torchpath = (Path(RootPath) / "torch").resolve()
    try:
        subscript = arg[0]
        p, ss = osp.split(subscript)
        torchpath = Path(osp.join(torchpath, p))
        if ss.endswith(".py"):
            ss = ss[:-3]
        logger.info(f"{ss}, {torchpath}")
        arg = arg[1:]
        load_scripts(ss, arg, torchpath)
    except Exception as e:
        logger.warning("Error running scripts: ", e) 

def main(**otherOpt):
    from argparse import ArgumentParser
    parser = ArgumentParser(prog='bffacility', description="Usage: bffacility <subcommand> [args]")
    parser.add_argument('subcmd', type=str, nargs="?", help=tr('type sub-command to exec certain function'))
    parser.add_argument('-V', action="version", help=tr(f'show version: {__version__}'), version=f'%(prog)s {__version__}')

    sys.argv.pop(0)  # remove script name
    args = vars(parser.parse_args(sys.argv[:1]))
    cmd = args["subcmd"]
    
    arg = sys.argv[1:]
    
    # logger.info(sys.stdout)

    if cmd == 'tray':
        from .win_tray import main as tray
        tray(arg)
        return
    elif cmd == 'pri':
        loadPriScript(arg)
    elif cmd == 'torch':
        loadTorch(arg)
    elif cmd in availableCmds:
        load_scripts(availableCmds[cmd], arg, **otherOpt)
    else:
        onError(parser)

# if __name__ == "__main__":
#     main()


import unittest
from bffacilities.socketserver import TcpSocketServer, SocketClient, logger
from bffacilities.utils import changeLoggerLevel
import logging
changeLoggerLevel(logger, logging.DEBUG)
import time
import threading
import asyncio

__version__ = "0.0.10"

async def createClient(id, address, waitCount):
    client = SocketClient.createTcpClient(address)
    def parse(x, c):
        print(f"{client} , {x}")
    client.setParser(parse, testJson=False)
    try:
        client.connect()
        data = f"{{'name': '{id}'}}"
        i = 0
        while i < waitCount:
            client.sock.send(data.encode())
            await asyncio.sleep(1) 
            # client.handle_receive()
            i += 1
    except (KeyboardInterrupt, SystemExit):
        pass
    except Exception as e:
        print("Client exception:", e)
    client.sock.close()
    print("Client Done ", id)
async def main(args):
    tasks = []
    for i in range(args["count"]):
        tasks.append(createClient(i, (args["host"], args["port"]), args["wait"]))
    print("start")
    try:
        await asyncio.gather(*tasks)
    except (KeyboardInterrupt, SystemExit):
        pass
    

if __name__ == "__main__":
    import argparse, sys
    parser = argparse.ArgumentParser("TestTcpSocketClient")
    parser.add_argument("--port", type=int, default=54134, help="set port")
    parser.add_argument("--host", type=str, default="127.0.0.1", help="set host")
    parser.add_argument("--count", type=int, default=10, help="parallel count")
    parser.add_argument("--wait", type=int, default=10, help="parallel count")
    args = vars(parser.parse_args(sys.argv[1:]))
    
    loop = asyncio.new_event_loop()
    loop.run_until_complete(main(args))




import json
import cv2 as cv
import os
import os.path as osp
from tqdm import tqdm
from PIL import Image
import base64
import numpy as np
from labelme.label_file import LabelFile, PY2, QT4, utils, io
import shutil as sh

__version__ = "0.1.0"

def encodeImage(imarr, ext=".jpg"):
    """@See Also LabelFile.load_image_file

    Arguments:
        imarr(np.ndarray | PIL.Image | str)

    Return:
        (str) base64 encoded image
    """
    imtype = type(imarr)
    if isinstance(imarr, Image.Image):
        pass
    elif imtype == np.ndarray:
        imarr = Image.fromarray(imarr)
    elif imtype == str:
        imarr = PIL.Image.open(imarr)
    else:
        raise ValueError("Unsupported Type: ", imtype)
    # apply orientation to image according to exif
    image_pil = utils.apply_exif_orientation(imarr)

    with io.BytesIO() as f:
        # ext = osp.splitext(filename)[1].lower()
        if PY2 and QT4:
            format = "PNG"
        elif ext in [".jpg", ".jpeg"]:
            format = "JPEG"
        else:
            format = "PNG"
        image_pil.save(f, format=format)
        f.seek(0)
        # return f.read()
        imageData = f.read()
    return base64.b64encode(imageData).decode("utf-8")


def calcY(line, x):
    return line[0] * x + line[1]

import copy
from PIL import ImageEnhance

class LabelmeModifier(object):
    """
    用于旋转、缩放 Labelme 已经标记好的图片

    Arguments:
        root(str): root directory that contains the raw image and json file
        generated by `labelme`
        dst_dir(str)
    """

    def __init__(self, root, dst_dir, image_suffix=".jpg", debug=False, **kwargs):
        self.root = root
        self.dst_dir = dst_dir
        self.debug = debug
        self.image_suffix = image_suffix
        files = [f for f in sorted(os.listdir(root))]
        self.images = [file for file in files if image_suffix in file]
        jsonfiles = [file for file in files if ".json" in file ]
        self.names = [img[: img.rfind(image_suffix)] for img in self.images]

        print(f"Found files at {root} images ({image_suffix}): {len(self.images)}, jsons: {len(jsonfiles)}")
        
        os.makedirs(self.dst_dir, exist_ok=True)


    def readJsonFile(self, file):
        loc = osp.join(self.root, file)
        if not osp.exists(loc):
            return None
        try:
            with open(loc, "r") as f:
                content = json.loads(f.read())
        except Exception as e:
            print(e)
            raise ValueError(f"Exception for {file}: ")
        return content


    def _resize(self, name, size, jsoncontent):
        """
        Note: **imageData in jsoncontent is not modified, it should be modified by the last step.**

        Arguments:
            name(str): filename stem without path location
            size(tuple): new size that resized image will be
            jsoncontent: correspoinding labelme json file content

        Return:
            tuple: (newImage, jsoncontent)
        """
        # resize images only
        image = Image.open(osp.join(self.root, f"{name}{self.image_suffix}"))
        iw, ih = image.size
        wscale = size[0] / iw
        hscale = size[1] / ih
        newImage = image.resize(size)

        if jsoncontent is None:
            if self.debug:
                print(f"Not Found json file for {name}")
            return newImage, None
        jsoncontent["imageWidth"]  = size[0]
        jsoncontent["imageHeight"] = size[1]
        if self.debug:
            newImage.show(f"Resized Image {name} {wscale} {hscale}")
        shapes = jsoncontent["shapes"]
        for shape in shapes:
            for point in shape["points"]:
                point[0] *= wscale
                point[1] *= hscale
        
        if self.debug:
            print("Name: ", name)
        return newImage, jsoncontent

    def resize(self, size, saveImageData=False):
        """
        Collabrate with crop, If crop is need before resize, using cropResize instead

        Arguemengs:
            size(tuple): W X H
        """
        count = 0
        for name in tqdm(self.names):
            jsoncontent = self.readJsonFile(f"{name}.json")
            # print("JsonContent", jsoncontent)
            newImage, jsoncontent = self._resize(name, size, jsoncontent)
            savedLoc = osp.join(self.dst_dir, f"{name}{self.image_suffix}")
            newImage.save(savedLoc)
            if jsoncontent is not None:
                imageData = LabelFile.load_image_file(savedLoc)
                imageData = base64.b64encode(imageData).decode("utf-8")
                jsoncontent["imageData"] = imageData if saveImageData else None

                with open(osp.join(self.dst_dir, f"{name}.json"), "w") as f:
                    f.write(json.dumps(jsoncontent, indent=True))
            count += 1
            if self.debug:
                break

        print(f"Done for {count}")

    def _rotate(self, name, jsoncontent, degree, center, boarderValue, scale=1.0):
        """
        Arguments: 
            name(str): Image Name
            degree(float): degrees
            center(tuple): X,Y

        See torchutils.rotateImage
        """
        image = Image.open(osp.join(self.root, f"{name}{self.image_suffix}"))
        iw, ih = image.size    #    
        if center is None:
            center = (iw // 2, ih // 2 )
        imarr = np.asarray(image)
        
        M = cv.getRotationMatrix2D(center, degree, scale)
        rotated = cv.warpAffine(imarr, M, (iw, ih), borderValue=boarderValue)
        # if self.debug:
        #     cv.imshow("default", rotated)
        #     cv.waitKey(0)

        if jsoncontent is None:
            return imarr

        shapes = jsoncontent["shapes"]
        for shape in shapes:
            points = []
            # rotatepoints = cv.warpAffine(points, M, (iw, ih), borderValue=boarderValue)
            # print(rotatepoints.shape, points.shape)
            for point in shape["points"]:
                newPoint = np.dot(M[:,:2], np.asarray(point)) + M[:,2]
                # print("\n=======", newPoint, M[:, 2])
                # check point pos
                if newPoint[0] > iw: newPoint[0] = iw
                if newPoint[1] > ih: newPoint[0] = ih
                points.append([newPoint[0], newPoint[1]])
                # if self.debug: break
            # if self.debug: break
            shape["points"] = points
        return Image.fromarray(rotated)
        

    def rotate(self, degs, center=None, boarderValue = (0, 0, 0), scale=1.0, **kwargs):
        """
        Arguments:
            degs: list or iterable (float)
            center(tuple | None): if the all the image size are the same, it could be applied when rotating,
                if image size differes from each other, leave it unchanged to let program choose the center 
                of the image
        """
        count = 0
        if kwargs.get("start"):
            print("[Warning] api changed, using `degs` instead")
        for name in tqdm(self.names):
            orijsoncontent = self.readJsonFile(f"{name}.json")
            if orijsoncontent is None:
                continue
            # print("JsonContent", jsoncontent)
            rc = 0
            for deg in degs:
                jsoncontent = copy.deepcopy(orijsoncontent) 
                newImage = self._rotate(name, jsoncontent, deg, center, boarderValue, scale=scale)
                newName = f"{name}-r{deg}-s{scale}-{rc:04}"
                savedLoc = osp.join(self.dst_dir, f"{newName}{self.image_suffix}")
                rc += 1
                newImage.save(savedLoc)
                if jsoncontent is not None:
                    imageData = LabelFile.load_image_file(savedLoc)
                    imageData = base64.b64encode(imageData).decode("utf-8")
                    jsoncontent["imageData"] = imageData

                    with open(osp.join(self.dst_dir, f"{newName}.json"), "w") as f:
                        f.write(json.dumps(jsoncontent, indent=True))
                count += 1

                if self.debug:
                    break
            if self.debug:
                break

        print(f"Done Rotation for {count}")

    def _shift(self, name, jsoncontent, xy, boarderValue):
        """
        Arguments: 
            name(str): Image Name
            degree(float): degrees
            center(tuple): X,Y

        See torchutils.rotateImage
        """
        image = Image.open(osp.join(self.root, f"{name}{self.image_suffix}"))
        iw, ih = image.size    #    

        if jsoncontent is None:
            return image

        image = image.rotate(0, translate=xy)
        shapes = jsoncontent["shapes"]
        for shape in shapes:
            points = []
            for point in shape["points"]:
                newPoint = [point[0] + xy[0], point[1] + xy[1]]
                # print("\n=======", newPoint, M[:, 2])
                # check point pos
                if newPoint[0] > iw: newPoint[0] = iw
                if newPoint[1] > ih: newPoint[0] = ih
                points.append([newPoint[0], newPoint[1]])
                # if self.debug: break
            # if self.debug: break
            shape["points"] = points
        return image
        
    def shift(self, maxpixel=10, step = 10, direction = (1, 1), boarderValue = (0, 0, 0)) :
        """
        pixel shift will be 10, 5 if max is 10 and step is 5
        direction is the vector that shift will be, for example (1, 1) 
        will shift the image to the up-right corner for pixel
        """
        count = 0
        assert direction[0] != 0 or direction[1] != 0
        for name in tqdm(self.names):
            orijsoncontent = self.readJsonFile(f"{name}.json")
            if orijsoncontent is None:
                continue
            # print("JsonContent", jsoncontent)
            rc = 0
            for pixel in range(maxpixel, 0, - step):
                if pixel <= 0: break
                xy = (direction[0] * pixel, direction[1] * pixel)
                jsoncontent = copy.deepcopy(orijsoncontent) 
                newImage = self._shift(name, jsoncontent, xy, boarderValue)
                newName = f"{name}-t{direction[0]}-{direction[1]}-{pixel}-{rc:04}"
                savedLoc = osp.join(self.dst_dir, f"{newName}{self.image_suffix}")
                rc += 1
                newImage.save(savedLoc)
                if jsoncontent is not None:
                    imageData = LabelFile.load_image_file(savedLoc)
                    imageData = base64.b64encode(imageData).decode("utf-8")
                    jsoncontent["imageData"] = imageData

                    with open(osp.join(self.dst_dir, f"{newName}.json"), "w") as f:
                        f.write(json.dumps(jsoncontent, indent=True))
                count += 1

                if self.debug:
                    break
            if self.debug:
                break

        print(f"Done Shift for {count}")

    def cropResize(self, leftTop, cropSize, newSize = None):
        """
        Arguements:
            leftTop(tuple): (X, Y)

            cropSize(tuple): W X H, croped size.  
                (cropSize + leftTop) should be less than original_size at W & H, 
                if not, (cropSize + leftTop) will reach the original images rightBottom

            newSize(tuple): W X H, rescaled size after cropped, if it's None, then it 
                won't be resized
        """
        raise NotImplementedError("Not Impletmented Yet")
        
    def copy(self):
        for name in tqdm(self.names):
            # imane
            imname = f"{name}{self.image_suffix}"
            sh.copy(osp.join(self.root, imname), osp.join(self.dst_dir, imname))
            jsonname = f"{name}.json"
            sh.copy(osp.join(self.root, jsonname), osp.join(self.dst_dir, jsonname))

    def output(self, destdir):
        """
        Arguments:
            destdir(str): dest directory
        """
        pass

        
    def polyfitSlope(self):
        for jsonfile in tqdm(jsonfiles):
        #     print(jsonfile)
            with open(src/jsonfile, "r") as f:
                content = json.loads(f.read())
            # print(content)
            shapes = content['shapes']
            imagePath = content['imagePath']
            occipitalSlope = None
            sphenoidbone = None
            newShapes = []
            for i, shape in enumerate(shapes):
                if shape['label'] == 'SphenoidBone':
                    sphenoidbone = shape
                elif shape['label'] == labelType:
                    occipitalSlope = shape
                    continue
                newShapes.append(shape)
                
            shapes = newShapes
                    
        #     if occipitalSlope is not None:
        #         continue
            imdata = cv.imread(str(src/ imagePath), 0)
        #     flatcontour = np.asarray(sphenoidbone["points"], dtype=int)
        #     contourLen = len(flatcontour)
            # print(flatcontour)
        #     maxy = (np.argmax(flatcontour[:, 1]))
        #     maxy = (maxy + 1) % len(flatcontour)
        #     if (maxy + 5) < len(flatcontour):
        #         contour = flatcontour[maxy:maxy+5]
        #     else:
        #         contour = flatcontour[maxy:]
        #         con = contour.shape[0]
        # #         print("length", con)
        #         contour = np.concatenate((contour, flatcontour[: 5-con]))
        #     line = np.polyfit(contour[:,0].reshape(-1), contour[:,1].reshape(-1), 1)
        #     x1 = float(contour[0, 0])
        #     x2 = float(contour[-1, 0] + 20)
        #     y1 = float(calcY(line, x1))
        #     y2 = float(calcY(line, x2))
        #     print([[x1, y1], [x2, y2]])
            # cv.line(imdata, (x1, y1), (x2, y2),(0,255,0), 2) 
            # plt.figure(figsize=(10, 8 ), dpi=100)
            # plt.imshow(imdata, cmap='gray', vmin=0, vmax=255)
        #     occipitalSlope = {"label": labelType, "points": [[x1, y1], [x2, y2], [x2, y2+1], [x1, y1+1]], "gorup_id": None, "shape_type": "polygon", "flags": {}}
        #     shapes.append(occipitalSlope)
        #     print([s["label"] for s in shapes], len(shapes))
        #     break
            content["shapes"] = shapes
            try:
                data = json.dumps(content, indent=True)
                
                with open(src/jsonfile, "w") as f:
                    f.write(data)

            except Exception as e:
                    print(e)

class LabelmeLabelModifier(LabelmeModifier):
    """专用于修改 Label 名称或调整 label 数量的
    """
    def removeLabels(self, labels):
        """
        Remove Labels if exits
        Arugments:
            labels(list(str))
        """
        assert type(labels) is list
        r_labels = {}
        for label in labels:
            r_labels[label] = { "num": 0 }
        self._renameShapes(r_labels)

   
    def _renameShapes(self, labelShapes):
        """Rename one shape at one time,

        Arguments:
            labelShapes(dict): 
            ```
            {
                "old_label_name": {
                    "label": "new_label_name", // empty for not change
                    "num": 1000, // default 1000,
                    "shape_idx": False, // default False
                }
            }
            ```
        """
        count = 0

        copyImage = (self.root != self.dst_dir)

        for name in tqdm(self.names):
            orijsoncontent = self.readJsonFile(f"{name}.json")
            if orijsoncontent is None:
                continue
            new_shapes = []
            reserve_count = 0
            # iterate over original json content
            for shape in orijsoncontent["shapes"]:
                # shape is a dict that contains label
                # print("====\n", shape, "====\n")
                oldName = shape["label"] 
                if oldName not in labelShapes:
                    # 如果找不到需要修改的label，默认保留旧 label
                    new_shapes.append(shape)
                    continue
                prop = labelShapes[oldName]
                # 默认使用旧名称
                new_label_name = prop.get("label", oldName)
                shape_idx = prop.get("shape_idx", False)
                num_reserved = prop.get("num", 1000)
                # if count not meet, reserve it
                if reserve_count < num_reserved:
                    nshape = copy.deepcopy(shape)
                    # 避免重复排序
                    try: 
                        d = new_label_name[new_label_name.rfind("-")+1:]
                        d = int(d)
                        shape_idx = False
                    except:
                        pass
                    if shape_idx:
                        nshape["label"] = new_label_name + f"-{reserve_count}"
                    else:
                        nshape["label"] = new_label_name
                    new_shapes.append(nshape)
                    reserve_count += 1
            # print(new_shapes)
            orijsoncontent["shapes"] = list(sorted(new_shapes, key=lambda x: x['label']))
            with open(osp.join(self.dst_dir, f"{name}.json"), "w") as f:
                f.write(json.dumps(orijsoncontent, indent=True))
            if copyImage:
                sh.copy(osp.join(self.root, f"{name}{self.image_suffix}"), self.dst_dir)
            count += 1
        print(f"Done rename shapes: {count}")

    def renameShapes(self, shapename, new_shapename=None, num_reserved=None, shape_idxes=None):
        """
        Arguments:
            shapename (list(str) ) | dict 
            new_shapename (list(str))
            num_reserved  list(Number) max_number of instance reserved, default 1000
            shape_idxes list(Boolean) whether to append idxes for different instance in one Picture
            
        Example:
        
        use one dict:

        ```py
        lm = LabelmeModifier(src_dir, dst_dir, image_suffix=".png")
        labels = {
            "Person": {
                "label": "Human",
                "num": 100, 
                "shape_idx": True,
            }
        }
        lm.renameShapes(labels)
        ```

        or use multiple arrays instead:

        ```py
        lm = LabelmeModifier(src_dir, dst_dir, image_suffix=".png")
        ori_label = ["Person"]
        new_label = ["Human"]
        nums = [100]
        lm.renameShapes(ori_label, new_label, nums)
        ```

        """
        stype = type(shapename)
        if stype == str:
            raise ValueError("String is not supported anymore. Please Use List instead.")
        if stype == dict:

            self._renameShapes(shapename)
        else:
            assert stype is list, " shapename must be either List or Dict"

            if new_shapename is None:
                new_shapename = shapename
            slen = len(shapename)
            if shape_idxes is None:
                shape_idxes = [False for _ in range(slen)]
            if num_reserved is None:
                num_reserved = [1000 for _ in range(slen)]
            assert len(new_shapename) == len(shapename) == len(shape_idxes) == len(num_reserved)
            shapes = {}
            for idx, name in enumerate(shapename):
                shapes[name] = {
                    "label": new_shapename[idx],
                    "num": num_reserved[idx],
                    "shape_idx": shape_idxes[idx]
                }
            self._renameShapes(shapes)
        


def moveImages(src, dst):
    """move Images from src dir with json files not moved
    """
    jsons = []
    imgs = []
    for root, dirs, files in os.walk(src):
        for file in files:
            if file.endswith(".json"):
                jsons.append(file[:file.rfind(".")])
            else:
                imgs.append(file)
    
    dst = str(dst)
    for file in imgs:
        filename = file[:file.rfind(".")]
        if filename in jsons:
            continue
        # print(filename)
        sh.move(str(src / file), dst)

from .labelme2voc import Labelme2Vocor
from .labelme2coco import Labelme2Cocoer



import torch
import torch.nn as nn
import torch.nn.functional as F   # 神经网络模块中的常用功能 
from torch.jit.annotations import List, Tuple, Dict, Optional
from torch.nn import Parameter
from collections import OrderedDict

class ResidualModule(nn.Module):
    """
    Residual Module whose in_channels(default) is 256, out_channels(default) is 256

    Arguments:
        input(int): in_channels
        batch(bool): whether to deploy batch normlization layer
        stride(int): stride will be applied in the first ConvLayer
    """
    def __init__(self, in_channels=256, out_channels=256, div=2, \
            batch=False, **kwargs
        ):
        super().__init__()
        # self.name = name

        # padding = kwargs.get('padding', 0)
        mid_channels = out_channels // div

        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1) # conv3
        self.conv4 = nn.Conv2d(out_channels, mid_channels, kernel_size=1)
        self.conv5 = nn.Conv2d(mid_channels,  mid_channels, kernel_size=3, padding=1) # padding with 0, to keep feature map size
        self.conv3 = nn.Conv2d(mid_channels, out_channels, kernel_size=1)  ## conv3
        self.batch = nn.BatchNorm2d(out_channels) if batch else None
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.relu(self.conv1(x))
        out1 = self.relu(self.conv4(x))
        out1 = self.relu(self.conv5(out1))
        out1 = self.relu(self.conv3(out1))
        x = x + out1
        if self.batch is not None:
            x = self.batch(x)
        return x

    def __repr__(self):
        return f"Resnet-{self.name}"


class AttentionResidualModule(nn.Module):
    expansion = 4

    def __init__(self, in_channels=256, out_channels=256, **kwargs):
        super().__init__()
        mid_channels = out_channels // 2
        self.conv0 = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self.conv4 = nn.Conv2d(out_channels, mid_channels, kernel_size=1, bias=False)
        self.conv5 = nn.Conv2d(mid_channels, mid_channels, kernel_size=3, stride=1,
                               padding=1, bias=False)
        self.conv3 = nn.Conv2d(mid_channels, out_channels, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)

        self.ca = ChannelAttention(out_channels)
        self.sa = SpatialAttention()
    def forward(self, x):
        x = self.relu(self.conv0(x))
        out = self.relu(self.conv4(x))

        out = self.relu(self.conv5(out))

        out = self.conv3(out)
        out = self.bn3(out)

        out = self.ca(out) * out
        out = self.sa(out) * out

        out += x
        out = self.relu(out)

        return out

class _HourGlassDownUnit(nn.Sequential):

    def __init__(self, in_channels, out_channels, block,\
        pool_kernel_size, pool_padding, **kwargs
        ):
        if pool_kernel_size is None:
            super().__init__(
                nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=3, stride=2, padding=1),
            )
        else:
            super().__init__(
                # nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=3, stride=1, padding=1),
                # ResidualModule(in_channels=in_channels, out_channels=out_channels, kernel_size=3, stride=1, padding=1),
                block(in_channels=in_channels, out_channels=out_channels, kernel_size=3, stride=1, padding=1),
                nn.MaxPool2d(pool_kernel_size, stride=2, padding=pool_padding),
            )
    
class _HourGlassUpUnit(nn.Sequential):

    def __init__(self, in_channels, out_channels, block, mode='nearest', **kwargs):
        super().__init__(
            # nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=3, stride=1, padding=1),
            # ResidualModule(in_channels=in_channels, out_channels=out_channels, kernel_size=3, stride=1, padding=1),
            block(in_channels=in_channels, out_channels=out_channels, kernel_size=3, stride=1, padding=1),
            nn.Upsample(scale_factor=2, mode=mode),
        )

class HourGlassModule(nn.Module):
    """Used for conv-deconv

    @Notes:
        Input feature map size should be greater than 32x32

    Arguments:
        in_channels: input feature map channels
        out_channels: output feature map channels

    Desc:
        channels will be always the same
    """

    def __init__(self, in_channels, out_channels, mid_channels=0, block=None, \
            debug=False, residual=False, **kwargs
        ):
        super().__init__()
        self.debug = debug
        if mid_channels == 0:
            mid_channels = out_channels
        # in_channels is 256, input feature map size is 32x32 / 64x64
        if block is None:
            block = ResidualModule if residual else nn.Conv2d
        kernel_size = (2, 2)
        padding = 0
        self.down_1 = _HourGlassDownUnit(in_channels, mid_channels, block=block, pool_kernel_size=kernel_size, pool_padding=padding, stride=2, batch=True)
        # 256 x 16 x 16  # 32 x 32
        self.down_2 = _HourGlassDownUnit(mid_channels, mid_channels, block=block, pool_kernel_size=kernel_size, pool_padding=padding, stride=2, batch=True)
        
        # 256 x 8 x 8
        self.down_3 = _HourGlassDownUnit(mid_channels, mid_channels, block=block, pool_kernel_size=kernel_size, pool_padding=padding, stride=2, batch=True)
        # 256 x 4 x 4   
        self.down_4 = _HourGlassDownUnit(mid_channels, mid_channels*2, block=block, pool_kernel_size=kernel_size, pool_padding=padding, stride=2, batch=True)
        # 256 x 2 x 2
        ###############################
        # 256 x 2 x2
        mode = kwargs.get('mode', 'nearest')
        self.up_4 = _HourGlassUpUnit(mid_channels*2, mid_channels, block=block, mode=mode)
        # 256 x 4 x 4
        self.up_3 = _HourGlassUpUnit(mid_channels, mid_channels, block=block, mode=mode)
        # 256 x 8 x 8
        self.up_2 = _HourGlassUpUnit(mid_channels, mid_channels, block=block, mode=mode)
        # 256 x 16 x 16
        self.up_1 = _HourGlassUpUnit(mid_channels, out_channels, block=block, mode=mode)
        # 256 x 32 x 32

    def forward(self, x):
        # print(x.shape)
        # 64 x 64
        dout1 = self.down_1(x)
        # 32 x 32
        dout2 = self.down_2(dout1)
        # 16 x 16
        dout3 = self.down_3(dout2)
        # 8 x 8
        dout4 = self.down_4(dout3)
        # 4 x 4

        # print(dout)
        out4 = F.relu(self.up_4(dout4))# + dout3)  # s3
        # 8
        out3 = F.relu(self.up_3(out4) + dout2) # s2
        # 16
        out2 = F.relu(self.up_2(out3) + dout1) # s1
        # 32
        out1 = F.relu(self.up_1(out2)) # + x) # according to paper HourGlass, x is not added
        # out1 = F.relu(self.up_1(out2) + x) # according to paper HourGlass, x is not added

        # 64
        if self.debug:
            return OrderedDict([
                ['0', out1], 
                ['1', out1], ['2', out2], ['3', out3],
                ['4', out4], ['5', dout4], ['6', dout3], ['7', dout2],
                # ['s3', s3], ['s2', s2], ['s1', s1]
            ])
        return out1

class HourGlassWrapper(nn.Module):
    """
    If last, the keypoint output will be returned instead of OrderedDict
    """
    def __init__(self, in_channels, out_channels, last=False, kp_output=True, **kwargs):
        super().__init__()
        self.last = last
        self.kp_output = True
        self.convDeconv = HourGlassModule(in_channels=in_channels, out_channels=in_channels, usePool=True, **kwargs)
        # self.res = ResidualModule(in_channels=in_channels, out_channels=in_channels, kernel_size=1)
        self.res = AttentionResidualModule(in_channels=in_channels, out_channels=in_channels, kernel_size=1)
        self.kpconv1 = nn.Conv2d(in_channels, out_channels, 1) # conv2
        if not self.last:
            self.conv = nn.Conv2d(in_channels, in_channels, 1) # conv3
            self.dconv2 = nn.Conv2d(out_channels, in_channels, 1) # conv3

    def forward(self, x, converge=True):
        uout1 = self.convDeconv(x)
        uout1 = self.res(uout1)

        if not self.kp_output:
            uout2 = F.relu(self.conv(uout1))
            return x + uout2

        out_kp1 = F.relu(self.kpconv1(uout1))
        # 4 x 32 x 32
        if not self.last:
            uout2 = F.relu(self.conv(uout1))
            if converge:
                dout2 = F.relu(self.dconv2(out_kp1))
                return OrderedDict([
                    ['0', x + uout2 + dout2], ['1', out_kp1]
                ])
            else:
                return OrderedDict([
                    ['0', x + uout2], ['1', out_kp1]
                ])
        return out_kp1

############## CapsNet ##################
import torch
import torch.nn as nn
import torch.nn.functional as F

class PrimaryCaps(nn.Module):
    '''
    The `PrimaryCaps` layer consists of 32 capsule units. Each unit takes
    the output of the `Conv1` layer, which is a `[256, 20, 20]` feature
    tensor (omitting `batch_size`), and performs a 2D convolution with 8
    output channels, kernel size 9 and stride 2, thus outputing a [8, 6, 6]
    tensor. In other words, you can see these 32 capsules as 32 paralleled 2D
    convolutional layers. Then we concatenate these 32 capsules' outputs and
    flatten them into a tensor of size `[1152, 8]`, representing 1152 8D
    vectors, and send it to the next layer `DigitCaps`.

    As indicated in Section 4, Page 4 in the paper, *One can see PrimaryCaps
    as a Convolution layer with Eq.1 as its block non-linearity.*, outputs of
    the `PrimaryCaps` layer are squashed before being passed to the next layer.

    Reference: Section 4, Fig. 1
    '''

    def __init__(self, in_resolution=64):
        '''
        We build 8 capsule units in the `PrimaryCaps` layer, each can be
        seen as a 2D convolution layer.
        '''
        super(PrimaryCaps, self).__init__()
        self.resolution1 = in_resolution
        self.resolution2 = self.resolution1 // 2 - 4

        num_caps = 32
        self.resolution3 = (self.resolution2 ** 2) * num_caps

        self.capsules = nn.ModuleList([
            nn.Sequential(
                nn.Conv2d(in_channels=256, out_channels=16, kernel_size=5, stride=2),
                nn.Conv2d(in_channels=16, out_channels=8, kernel_size=3, stride=1, padding=1),
                nn.Conv2d(in_channels=8, out_channels=8, kernel_size=3, stride=1, padding=0),
            )
            
            for i in range(num_caps)
        ])

    def forward(self, x):
        '''
        Each capsule outputs a [batch_size, 8, 6, 6] tensor, we need to
        flatten and concatenate them into a [batch_size, 8, 6*6, 32] size
        tensor and flatten and transpose into `u` [batch_size, 1152, 8], 
        where each [batch_size, 1152, 1] size tensor is the `u_i` in Eq.2. 

        #### Dimension transformation in this layer(ignoring `batch_size`):
        [256, 20, 20] --> [8, 6, 6] x 32 capsules --> [1152, 8]

        Note: `u_i` is one [1, 8] in the final [1152, 8] output, thus there are
        1152 `u_i`s.
        '''
        batch_size = x.size(0)

        u = []
        for i in range(32):
            # Input: [batch_size, 256, 20, 20]
            assert x.shape[-2:] == (self.resolution1, self.resolution1), f"{x.shape}, {self.resolution1}"

            u_i = self.capsules[i](x)
            assert u_i.shape[-3:] == (8, self.resolution2, self.resolution2)
            # u_i: [batch_size, 8, 6, 6]
            u_i = u_i.view(batch_size, 8, -1, 1) # 24 
            # u_i: [batch_size, 8, 36] # 576
            u.append(u_i)

        # u: [batch_size, 8, 36, 1] x 32
        u = torch.cat(u, dim=3)
        # u: [batch_size, 8, 36, 32]
        u = u.view(batch_size, 8, -1) # 576 * 32 == 
        # u: [batch_size, 8, 1152]
        u = torch.transpose(u, 1, 2)
        # u: [batch_size, 1152, 8]
        assert u.data.shape[-2:] == (self.resolution3, 8)

        # Squash before output
        u_squashed = self.squash(u)

        return u_squashed

    def squash(self, u):
        '''
        Args:
            `u`: [batch_size, 1152, 8]

        Return:
            `u_squashed`: [batch_size, 1152, 8]

        In CapsNet, we use the squash function after the output of both 
        capsule layers. Squash functions can be seen as activating functions
        like sigmoid, but for capsule layers rather than traditional fully
        connected layers, as they squash vectors instead of scalars.

        v_j = (norm(s_j) ^ 2 / (1 + norm(s_j) ^ 2)) * (s_j / norm(s_j))

        Reference: Eq.1 in Section 2.
        '''
        batch_size = u.size(0)

        # u: [batch_size, 1152, 8]
        square = u ** 2

        # square_sum for u: [batch_size, 1152]
        square_sum = torch.sum(square, dim=2)

        # norm for u: [batch_size, 1152]
        norm = torch.sqrt(square_sum)

        # factor for u: [batch_size, 1152]
        factor = norm ** 2 / (norm * (1 + norm ** 2))

        # u_squashed: [batch_size, 1152, 8]
        u_squashed = factor.unsqueeze(2) * u
        assert u_squashed.shape[-2:] == (self.resolution3, 8)

        return u_squashed


class Decoder(nn.Module):
    '''
    The decoder network consists of 3 fully connected layers. For each
    [10, 16] output, we mask out the incorrect predictions, and send
    the [16,] vector to the decoder network to reconstruct a [784,] size
    image.

    Reference: Section 4.1, Fig. 2
    '''

    def __init__(self, out_resolution, num_classes=10):
        '''
        The decoder network consists of 3 fully connected layers, with
        512, 1024, 784 neurons each.
        '''
        super().__init__()
        self.num_classes = num_classes

        self.fc1 = nn.Linear(16, 512)
        self.fc2 = nn.Linear(512, 1024)
        self.out_resolution = out_resolution // 2
        self.final_resolution = out_resolution ** 2
        self.fc3 = nn.Linear(1024, self.out_resolution ** 2)
        self.up =  nn.UpsamplingBilinear2d(scale_factor=2)

    def forward(self, v, target):
        '''
        Args:
            `v`: [batch_size, 10, 16]
            `target`: [batch_size, 10]

        Return:
            `reconstruction`: [batch_size, 784]

        We send the outputs of the `DigitCaps` layer, which is a
        [batch_size, 10, 16] size tensor into the decoder network, and
        reconstruct a [batch_size, 784] size tensor representing the image.
        '''
        batch_size = target.size(0)

        target = target.type(torch.FloatTensor)
        # mask: [batch_size, 10, 16]
        mask = torch.stack([target for i in range(16)], dim=2)
        assert mask.shape[-2:] == (10, 16)
        if v.device != torch.device("cpu"):
            mask = mask.to(v.device)

        # v: [bath_size, 10, 16]
        v_masked = mask * v
        v_masked = torch.sum(v_masked, dim=1)
        assert v_masked.shape[-1] == 16

        # Forward
        v = F.relu(self.fc1(v_masked))
        v = F.relu(self.fc2(v))
        reconstruction = torch.sigmoid(self.fc3(v))

        # assert reconstruction.shape[-1] == self.out_resolution
        reconstruction = reconstruction.view(batch_size, 1, self.out_resolution, self.out_resolution)
        reconstruction = self.up(reconstruction)
        return reconstruction # .view(batch_size, self.final_resolution)


class DigitCaps(nn.Module):
    '''
    The `DigitCaps` layer consists of 10 16D capsules. Compared to the traditional
    scalar output neurons in fully connected networks(FCN), the `DigitCaps` layer
    can be seen as an FCN with ten 16-dimensional output neurons, which we call
    these neurons "capsules".

    In this layer, we take the input `[1152, 8]` tensor `u` as 1152 [8,] vectors
    `u_i`, each `u_i` is a 8D output of the capsules from `PrimaryCaps` (see Eq.2
    in Section 2, Page 2) and sent to the 10 capsules. For each capsule, the tensor
    is first transformed by `W_ij`s into [1152, 16] size. Then we perform the Dynamic
    Routing algorithm to get the output `v_j` of size [16,]. As there are 10 capsules,
    the final output is [16, 10] size.

    #### Dimension transformation in this layer(ignoring `batch_size`):
    [1152, 8] --> [1152, 16] --> [1, 16] x 10 capsules --> [10, 16] output

    Note that in our codes we have vectorized these computations, so the dimensions
    above are just for understanding, actual dimensions of tensors are different.
    '''

    def __init__(self, routing, in_resolution, num_classes=10):
        '''
        There is only one parameter in this layer, `W` [1, 1152, 10, 16, 8], where
        every [8, 16] is a weight matrix W_ij in Eq.2, that is, there are 11520
        `W_ij`s in total.

        The the coupling coefficients `b` [64, 1152, 10, 1] is a temporary variable which
        does NOT belong to the layer's parameters. In other words, `b` is not updated
        by gradient back-propagations. Instead, we update `b` by Dynamic Routing
        in every forward propagation. See the docstring of `self.forward` for details.
        '''
        super(DigitCaps, self).__init__()
        self.routing = routing
        self.resolution1 = in_resolution
        self.num_classes = num_classes
        self.W = nn.Parameter(torch.randn(1, in_resolution, self.num_classes, 8, 16))

    def forward(self, u):
        '''
        Args:
            `u`: [batch_size, 1152, 8]
        Return:
            `v`: [batch_size, 10, 16]

        In this layer, we vectorize our computations by calling `W` and using
        `torch.matmul()`. Thus the full computaion steps are as follows.
            1. Expand `W` into batches and compute `u_hat` (Eq.2)
            2. Line 2: Initialize `b` into zeros
            3. Line 3: Start Routing for `r` iterations:
                1. Line 4: c = softmax(b)
                2. Line 5: s = sum(c * u_hat)
                3. Line 6: v = squash(s)
                4. Line 7: b += u_hat * v

        The coupling coefficients `b` can be seen as a kind of attention matrix
        in the attentional sequence-to-sequence networks, which is widely used in
        Neural Machine Translation systems. For tutorials on  attentional seq2seq
        models, see https://arxiv.org/abs/1703.01619 or
        http://pytorch.org/tutorials/intermediate/seq2seq_translation_tutorial.html

        Reference: Section 2, Procedure 1
        '''
        batch_size = u.size(0)

        # First, we need to expand the dimensions of `W` and `u` to compute `u_hat`
        assert u.shape[-2:] == (self.resolution1, 8)
        # u: [batch_size, 1152, 1, 1, 8]
        u = torch.unsqueeze(u, dim=2)
        u = torch.unsqueeze(u, dim=2)
        # Now we compute u_hat in Eq.2
        # u_hat: [batch_size, 1152, 10, 16]
        u_hat = torch.matmul(u, self.W).squeeze()

        # Line 2: Initialize b into zeros
        # b: [batch_size, 1152, 10, 1]
        b = torch.zeros(batch_size, self.resolution1, self.num_classes, 1)
        if b.device != u.device:
            b = b.to(u.device)

        # Start Routing
        for r in range(self.routing):
            # Line 4: c_i = softmax(b_i)
            # c: [b, 1152, 10, 1]
            c = F.softmax(b, dim=2)
            assert c.shape[-3:] == (self.resolution1, self.num_classes, 1)

            # Line 5: s_j = sum_i(c_ij * u_hat_j|i)
            # u_hat: [batch_size, 1152, 10, 16]
            # s: [batch_size, 10, 16]
            s = torch.sum(u_hat * c, dim=1)

            # Line 6: v_j = squash(s_j)
            # v: [batch_size, 10, 16]
            v = self.squash(s)
            assert v.shape[-2:] == ( self.num_classes, 16)

            # Line 7: b_ij += u_hat * v_j
            # u_hat: [batch_size, 1152, 10, 16]
            # v: [batch_size, 10, 16]
            # a: [batch_size, 10, 1152, 16]
            a = u_hat * v.unsqueeze(1)
            # b: [batch_size, 1152, 10, 1]
            b = b + torch.sum(a, dim=3, keepdim=True)

        return v

    def squash(self, s):
        '''
        Args:
            `s`: [batch_size, 10, 16]

        v_j = (norm(s_j) ^ 2 / (1 + norm(s_j) ^ 2)) * (s_j / norm(s_j))

        Reference: Eq.1 in Section 2.
        '''
        batch_size = s.size(0)

        # s: [batch_size, 10, 16]
        square = s ** 2

        # square_sum for v: [batch_size, 10]
        square_sum = torch.sum(square, dim=2)

        # norm for v: [batch_size, 10]
        norm = torch.sqrt(square_sum)

        # factor for v: [batch_size, 10]
        factor = norm ** 2 / (norm * (1 + norm ** 2))

        # v: [batch_size, 10, 16]
        v = factor.unsqueeze(2) * s
        assert v.shape[-2:] == (10, 16)

        return v


class CapsNet(nn.Module):

    def __init__(self, routing, num_classes = 10, **kwargs):
        '''
        The CapsNet consists of 3 layers: `Conv1`, `PrimaryCaps`, `DigitCaps`.`Conv1`
        is an ordinary 2D convolutional layer with 9x9 kernels, stride 2, 256 output
        channels, and ReLU activations. `PrimaryCaps` and `DigitCaps` are two capsule
        layers with Dynamic Routing between them. For further details of these two
        layers, see the docstrings of their classes. For each [1, 28, 28] input image,
        CapsNet outputs a [16, 10] tensor, representing the 16-dimensional output
        vector from 10 digit capsules.

        Reference: Section 4, Figure 1
        '''
        super().__init__()

        self.resolution1 = 28
        self.reconstruct_factor = 0.0005
        self.num_classes = num_classes
        # self.premodel =  nn.Sequential(
        #     nn.Conv2d(in_channels=1, out_channels=8, kernel_size=5, stride=2, padding=2), # 64
        #     nn.Conv2d(in_channels=8, out_channels=8, kernel_size=5, stride=2, padding=2), # 32
        #     nn.Conv2d(in_channels=8, out_channels=1, kernel_size=3, stride=1, padding=0),
        #     nn.Conv2d(in_channels=1, out_channels=1, kernel_size=3, stride=1, padding=0)
        # )
        
        self.premodel =  nn.Sequential(
            nn.Conv2d(in_channels=1,  out_channels=8,   kernel_size=5, stride=2, padding=0), # 62
            nn.Conv2d(in_channels=8,  out_channels=16,  kernel_size=3, stride=1, padding=0),  # 60
            nn.Conv2d(in_channels=16, out_channels=32,  kernel_size=5, stride=2, padding=0), # 28
            nn.Conv2d(in_channels=32, out_channels=256, kernel_size=3, stride=1, padding=1), # 28
        )
        # self.Conv1 = nn.Conv2d(in_channels=1, out_channels=256, kernel_size=9) # input_dim - 8
        self.primaryCaps = PrimaryCaps(in_resolution=self.resolution1)
        self.DigitCaps = DigitCaps(routing, self.primaryCaps.resolution3, num_classes=self.num_classes)

        self.Decoder = Decoder(self.resolution1, num_classes=self.num_classes)

    def forward(self, x):
        '''
        Args:
            `x`: [batch_size, 1, 28, 28] MNIST samples
        
        Return:
            `v`: [batch_size, 10, 16] CapsNet outputs, 16D prediction vectors of
                10 digit capsules

        The dimension transformation procedure of an input tensor in each layer:
            0. Input: [batch_size, 1, 28, 28] -->
            1. `Conv1` --> [batch_size, 256, 20, 20] --> 
            2. `PrimaryCaps` --> [batch_size, 8, 6, 6] x 32 capsules --> 
            3. Flatten, concatenate, squash --> [batch_size, 1152, 8] -->
            4. `W_ij`s and `DigitCaps` --> [batch_size, 16, 10] -->
            5. Length of 10 capsules --> [batch_size, 10] output probabilities
        '''
        # Input: [batch_size, 1, 28, 28]
        x = F.relu(self.premodel(x))
        # x = F.relu(self.Conv1(x))
        # PrimaryCaps input: [batch_size, 256, 20, 20]
        u = self.primaryCaps(x)
        # PrimaryCaps output u: [batch_size, 1152, 8] # 18432
        v = self.DigitCaps(u)
        # DigitCaps output v: [batsh_size, 10, 16]
        return v

    def marginal_loss(self, v, target, l=0.5):
        '''
        Args:
            `v`: [batch_size, 10, 16]
            `target`: [batch_size, 10]
            `l`: Scalar, lambda for down-weighing the loss for absent digit classes

        Return:
            `marginal_loss`: Scalar
        
        L_c = T_c * max(0, m_plus - norm(v_c)) ^ 2 + lambda * (1 - T_c) * max(0, norm(v_c) - m_minus) ^2
        
        Reference: Eq.4 in Section 3.
        '''
        batch_size = v.size(0)

        square = v ** 2
        square_sum = torch.sum(square, dim=2)
        # norm: [batch_size, 10]
        norm = torch.sqrt(square_sum)
        assert norm.size() == torch.Size([batch_size, 10])

        # The two T_c in Eq.4
        T_c = target.type(torch.FloatTensor)
        zeros = torch.zeros(norm.size())
        # Use GPU if available
        if v.device != torch.device("cpu"):
            zeros = zeros.to(v.device)
            T_c = T_c.to(v.device)

        # Eq.4
        marginal_loss = T_c * (torch.max(zeros, 0.9 - norm) ** 2) + \
            (1 - T_c) * l * (torch.max(zeros, norm - 0.1) ** 2)
        marginal_loss = torch.sum(marginal_loss)

        return marginal_loss

    def reconstruction_loss(self, reconstruction, image):
        '''
        Args:
            `reconstruction`: [batch_size, 784] Decoder outputs of images
            `image`: [batch_size, 1, 28, 28] MNIST samples

        Return:
            `reconstruction_loss`: Scalar Variable

        The reconstruction loss is measured by a squared differences
        between the reconstruction and the original image. 

        Reference: Section 4.1
        '''
        batch_size = image.size(0)
        # image: [batch_size, 784]
        image = F.interpolate(image, size=self.resolution1) # added by bf.
        # print(image.shape)
        image = image.unsqueeze(1)

        assert image.shape[-2:] == (self.resolution1, self.resolution1), image.shape
        
        # Scalar Variable
        reconstruction_loss = torch.sum((reconstruction - image) ** 2)
        return reconstruction_loss

    def loss(self, v, target, image):
        '''
        Args:
            `v`: [batch_size, 10, 16] CapsNet outputs
            `target`: [batch_size, 10] One-hot MNIST labels
            `image`: [batch_size, 1, 28, 28] MNIST samples

        Return:
            `L`: Scalar Variable, total loss
            `marginal_loss`: Scalar Variable
            `reconstruction_loss`: Scalar Variable

        The reconstruction loss is scaled down by 5e-4, serving as a
        regularization method.

        Reference: Section 4.1
        '''
        batch_size = image.size(0)

        marginal_loss = self.marginal_loss(v, target)

        # Get reconstructions from the decoder network
        reconstruction = self.Decoder(v, target)
        reconstruction_loss = self.reconstruction_loss(reconstruction, image)

        # Scalar Variable
        loss = (marginal_loss + self.reconstruct_factor * reconstruction_loss) / batch_size

        return loss, marginal_loss / batch_size, reconstruction_loss / batch_size
#######


# Resenet cbam

import torch
import torch.nn as nn
import math
import torch.utils.model_zoo as model_zoo

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}


def conv3x3(in_planes, out_planes, stride=1):
    "3x3 convolution with padding"
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)

class ChannelAttention(nn.Module):
    def __init__(self, in_planes, ratio=16):
        super(ChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)

        self.conv1   = nn.Conv2d(in_planes, in_planes // ratio, 1, bias=False)
        self.relu1 = nn.ReLU()
        self.conv2   = nn.Conv2d(in_planes // ratio, in_planes, 1, bias=False)

        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = self.conv2(self.relu1(self.conv1(self.avg_pool(x))))
        max_out = self.conv2(self.relu1(self.conv1(self.max_pool(x))))
        out = avg_out + max_out
        return self.sigmoid(out)

class SpatialAttention(nn.Module):
    def __init__(self, kernel_size=7):
        super(SpatialAttention, self).__init__()

        assert kernel_size in (3, 7), 'kernel size must be 3 or 7'
        padding = 3 if kernel_size == 7 else 1

        self.conv1 = nn.Conv2d(2, 1, kernel_size, padding=padding, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        """
            x : B C H W
        """
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        x = torch.cat([avg_out, max_out], dim=1)
        x = self.conv1(x)
        return self.sigmoid(x)

class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)

        self.ca = ChannelAttention(planes)
        self.sa = SpatialAttention()

        self.downsample = downsample
        # self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        out = self.ca(out) * out
        out = self.sa(out) * out

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)

        self.ca = ChannelAttention(planes * 4)
        self.sa = SpatialAttention()

        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out = self.ca(out) * out
        out = self.sa(out) * out

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, block, layers, num_classes=1000):
        self.inplanes = 64
        super(ResNet, self).__init__()
        # self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
        #                        bias=False)
        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=5, stride=2, padding=2,
                               bias=False),
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1,
                               bias=False)
        )
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        # print(x.shape)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x


def resnet18_cbam(pretrained=False, **kwargs):
    """Constructs a ResNet-18 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    if pretrained:
        pretrained_state_dict = model_zoo.load_url(model_urls['resnet18'])
        now_state_dict        = model.state_dict()
        now_state_dict.update(pretrained_state_dict)
        model.load_state_dict(now_state_dict)
    return model


def resnet34_cbam(pretrained=False, **kwargs):
    """Constructs a ResNet-34 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    if pretrained:
        pretrained_state_dict = model_zoo.load_url(model_urls['resnet34'])
        now_state_dict        = model.state_dict()
        now_state_dict.update(pretrained_state_dict)
        try:
            model.load_state_dict(now_state_dict)
        except : pass
    return model


def resnet50_cbam(pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        pretrained_state_dict = model_zoo.load_url(model_urls['resnet50'])
        now_state_dict        = model.state_dict()
        now_state_dict.update(pretrained_state_dict)
        try:
            model.load_state_dict(now_state_dict, strict=False)
        except : pass
    return model


def resnet101_cbam(pretrained=False, **kwargs):
    """Constructs a ResNet-101 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        pretrained_state_dict = model_zoo.load_url(model_urls['resnet101'])
        now_state_dict        = model.state_dict()
        now_state_dict.update(pretrained_state_dict)
        model.load_state_dict(now_state_dict)
    return model


def resnet152_cbam(pretrained=False, **kwargs):
    """Constructs a ResNet-152 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    if pretrained:
        pretrained_state_dict = model_zoo.load_url(model_urls['resnet152'])
        now_state_dict        = model.state_dict()
        now_state_dict.update(pretrained_state_dict)
        model.load_state_dict(now_state_dict)
    return model



import RPi.GPIO as GPIO
import time
import random
import subprocess as sp
from datetime import datetime as dt
from bffacilities import createLogger
import os.path as osp
secs = random.randint(60, 200)
logger = createLogger("gpio", stream=True)

if osp.exists("/home/pi/rwfile.py"):
    p = sp.Popen(["python3", "/home/pi/rwfile.py"])

# GPIO.setmode(GPIO.BOARD)
GPIO.setmode(GPIO.BCM)
pin = 20
GPIO.setup(pin, GPIO.OUT)
logger.info(f"Secs: {secs}")
time.sleep(secs)
logger.info(f"Start gpio hight")
GPIO.output(pin, GPIO.HIGH)
time.sleep(30)
exit()
while True:
    GPIO.output(pin, GPIO.LOW)
    time.sleep(10)
    GPIO.output(pin, GPIO.HIGH)
    time.sleep(10)




import torch
from torch.utils.data import Dataset
from torchvision.datasets import CocoDetection
from pycocotools.coco import COCO
import os
import os.path as osp
from pycocotools.coco import COCO
from PIL import Image
import json

class CocoDataset(CocoDetection):
    """ If num_classes or reprLabel should be used, the Labels should be set first.

    Arguments:       
        transform (callable, optional): A function/transform that  takes in an PIL image
            and returns a transformed version. E.g, ``transforms.ToTensor``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        transforms (callable, optional): A function/transform that takes input sample and its target as entry
            and returns a transformed version.
    """
    Labels = []

    @staticmethod
    def num_classes():
        return len(CocoDataset.Labels)

    @staticmethod
    def reprLabel(idx):
        return CocoDataset.Labels[idx]

    def __init__(self, root, annFile = None, \
            # whether to apply mask , keypoint
            mask=True, keypoint=True, \
            transform=None, target_transform=None, transforms=None, \
            train=False, debug = False, num_keypoints=2, coco_style_kp=False,
            **kwargs):
        

        if annFile is None:
            annFile = osp.join(root, "annotations.json")
        # print(root, annFile)
        super().__init__(osp.join(root, "JPEGImages"), annFile, transform=transform, \
            target_transform=target_transform, transforms=transforms)
        self.root = str(root)
        self.isTrain = train
        self.num_keypoints = num_keypoints
        self.coco_style_kp = coco_style_kp
        self._mask = mask
        self._keypoint = keypoint
        self.debug = debug
        print("[DS] Dataset root: ", self.root, "Len: ", len(self.ids))
        print("[DS] Mask: ", self._mask, " , Keypoint: ", self._keypoint)
    
    def load_annotations(self, ann_file, catNms):
        """Load annotation from COCO style annotation file.

        Args:
            ann_file (str): Path of annotation file.

        Returns:
            list[dict]: Annotation info from COCO api.
        """

        self.coco = COCO(ann_file)
        # self.cat_ids = self.coco.getCatIds(cat_names=self.CLASSES)
        self.cat_ids = self.coco.getCatIds(
            catNms=catNms
            )
        cats = self.coco.loadCats(self.cat_ids)
        nms=[cat['name'] for cat in cats]
        print('COCO categories: {}\n'.format(' '.join(nms)))

        self.cat2label = {cat_id: i for i, cat_id in enumerate(self.cat_ids)}
        self.img_ids = self.coco.getImgIds(catIds=self.cat_ids)
        img_infos = []
        for i in self.img_ids:
            info = self.coco.loadImgs([i])[0]
            # info['filename'] = info['file_name']
            img_infos.append(info)
        return img_infos 

    def get_filename(self, index):
        with open(osp.join(self.root, "annotations.json"), "r") as f:
            content = json.load(f)["images"]
        for image in content:
            if image["id"] == index:
                return image["file_name"]
        return None

    def get_idx_by_name(self, name):
        with open(osp.join(self.root, "annotations.json"), "r") as f:
            content = json.load(f)["images"]
        idxes = []
        for image in content:
            file_name = image["file_name"]
            if name in file_name:
                idxes.append((image["id"], file_name))
        return idxes

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: Tuple (image, target). target is the object returned by ``coco.loadAnns``.
        """
        coco = self.coco
        img_id = self.ids[index]
        ann_ids = coco.getAnnIds(imgIds=img_id)
        target = coco.loadAnns(ann_ids)
        # print("DS: ", ann_ids, target)

        boxes =  []
        labels =  []
        areas =  []
        iscrowds =  []
        for t in target:
            label = t["category_id"]
            labels.append(label)
            iscrowds.append(0)
            x,y,w,h=t["bbox"]
            boxes.append([x,y,x+w,y+h])
            areas.append(t["area"]) # coco mask area is slightly different from torch
            # areas.append(w*h)  
        t_target = {
            "image_id": torch.tensor([index]),
            "boxes": torch.as_tensor(boxes, dtype=torch.float32),
            "labels": torch.as_tensor(labels, dtype=torch.int64),
            "area": torch.as_tensor(areas, dtype=torch.float32),
            "iscrowd": torch.as_tensor(iscrowds, dtype=torch.uint8),
        }
        
        if self._mask:       
            masks =  []
            for t in target:
                label = t["category_id"]
                masks.append(coco.annToMask(t))
                
            t_target["masks"] = torch.as_tensor(masks, dtype=torch.uint8)
            # end mask
        if self._keypoint:
            # N K 3
            keypoints = [] 
            # print("Load Anns: ", target, len(target))
            for t in target:
                kps = []
                # print(t, "========", "keypoints" in t)
                if "keypoints" not in t:
                    ## TODO detect how many keypoints there has
                    for _ in range(self.num_keypoints):
                        kps.append([0, 0, 0])
                    keypoints.append(kps)
                    continue
                kp = t["keypoints"]
                xs = kp[0::3]
                ys = kp[1::3]
                vs = kp[2::3]
                if not self.coco_style_kp :
                    for x, y, v in zip(xs, ys, vs):
                        if v >= 1:
                            kps.append([x, y, v-1])
                        else:
                            kps.append([x, y, 0])
                else:
                    for x, y, v in zip(xs, ys, vs):
                        kps.append([x, y, v])
                keypoints.append(kps)
            # print(keypoints)
            t_target["keypoints"] = torch.as_tensor(keypoints, dtype=torch.float32)
            # end keypoint
        path = coco.loadImgs(img_id)[0]['file_name']
        
        img = Image.open(osp.join(self.root, path)).convert('RGB')
        if self.transforms is not None:
            # print(t_target)
            assert t_target is not None
            img, t_target = self.transforms(img, t_target)
        # print(t_target)
        return img, t_target

import numpy as np

class MaskDataset(Dataset):
    """
    Only for mask prediction and voc-style dataset

    @Note If your dataset is labeled by `labelme`, you should consider convert
    it first or use LabelmeDataset.
    @See labelmescripts for how to convert dataset from labelme format to voc
    @Update 21.03.12

    """

    Labels = None
    @staticmethod
    def num_classes():
        return len(MaskDataset.Labels)

    @staticmethod
    def reprLabel(idx):
        """idx starts from 1,
        0 represents background
        """
        return MaskDataset.Labels[idx]

    def __init__(self, root, transforms=None, train=True, **kwargs):
        super().__init__()
        self.root = str(root)
        self.name = kwargs.get("name")
        self.transforms = transforms
        self.isTrain = train
        if self.isTrain:
            imgs = []
            for img in sorted(os.listdir(osp.join(root, "JPEGImages"))):
                if (not img.endswith(".json") and not osp.isdir(osp.join(root,img))):
                    imgs.append(img)
            self.imgs = imgs
            imgs = []
            for img in sorted(os.listdir(osp.join(root, "SegmentationClassPNG"))):
                if (not img.endswith(".json") and not osp.isdir(osp.join(root,img))):
                    imgs.append(img)
            self.masks = imgs
        else:
            imgs = []
            for img in sorted(os.listdir(root)):
                if (not img.endswith(".json") and not osp.isdir(osp.join(root,img))):
                    imgs.append(img)
            self.imgs = imgs

        print("TrainDataset Size: ", len(self.imgs))
        # print("Dataset prepared")

    def getMask(self, idx):
        mask_path = osp.join(self.root, "SegmentationClassPNG", self.masks[idx])
        mask = Image.open(mask_path)
        mask = np.array(mask)
        return mask

    def getTarget(self, idx, mask, to_tensor=True):
        """All values are numpy datatype, Not converted to Tensor
        """
        # instances are encoded as different colors
        obj_ids = np.unique(mask)
        # first id is the background, so remove it
        obj_ids = obj_ids[1:]
        masks = (mask == obj_ids[:, None, None])
        # get bounding box coordinates for each mask
        num_objs = len(obj_ids)
        boxes = []
        labels = []
        iscrowd = []
        for i in range(num_objs):
            pos = np.where(masks[i])
            label = np.max(mask[masks[i]])
            # print(obj_ids, label)
            xmin = np.min(pos[1])
            xmax = np.max(pos[1])
            ymin = np.min(pos[0])
            ymax = np.max(pos[0])
            boxes.append([xmin, ymin, xmax, ymax])
            labels.append(label)
            iscrowd.append(0)
        # print(boxes)

        target = {}

        # target["maskimg"] = mask
        if to_tensor:
            target["image_id"] = torch.tensor([idx])
            boxes = torch.as_tensor(boxes, dtype=torch.float32)
            target["boxes"] = boxes
            # two class currently
            target["labels"] = torch.tensor(labels, dtype=torch.int64)
            target["masks"] = torch.as_tensor(masks, dtype=torch.uint8)
            # target["area"] = torch.tensor(target["area"], dtype=torch.uint8)
            target["area"] = (boxes[:, 3] - boxes[:, 1]) * (boxes[:, 2] - boxes[:, 0])
            target["iscrowd"] = torch.as_tensor(iscrowd, dtype=torch.int64)
        else:
            target["boxes"] = boxes
            target["labels"] = labels
            target["masks"] = masks
            target["image_id"] = [idx]
            target["iscrowd"] = iscrowd

        return target

    def __getitem__(self, idx):
        if not self.isTrain:
            img_path = osp.join(self.root, self.imgs[idx])   
            img = Image.open(img_path)

            img = np.array(img)
            # img = torch.Tensor(img)
            if self.transforms is not None:
                img, _ = self.transforms(img)
            else:
                img = FT.to_tensor(img)
            return img, None

        mask_path = osp.join(self.root, "SegmentationClassPNG", self.masks[idx])
        mask = Image.open(mask_path)
        mask = np.array(mask)
        target = self.getTarget(idx, mask)

        img_path = osp.join(self.root, "JPEGImages", self.imgs[idx])
        img = Image.open(img_path).convert('L')
        img = np.array(img)

        if self.transforms is not None:
            img, target = self.transforms(img, target)
        else:
            img = FT.to_tensor(img)

        return img, target


    def __len__(self):
        return len(self.imgs)

    def __repr__(self):
        return f'<DS {self.name}>'
import json
class LabelmeDataset(Dataset):

    def num_classes():
        return 0
    def reprLabel(idx):
        """ids starts from 1
        0 represents background
        """
        return None

    def __init__(self, rootDir, debug = False, train=True, transforms=None, imgFormat=None, **kwargs):
        self._root = str(rootDir)
        self.name = kwargs.get("name")
        self.transforms = transforms
        self.isTrain = train
        if imgFormat is None:
            imgFormat = [ ".jpg", ".png" ]
        
        assert type(imgFormat) is list

        # save all images json file name
        if self.isTrain:
            imgs = []
            jsons = []
            json_imgs = []
            for filename in sorted(os.listdir(self._root)):
                if osp.isdir(osp.join(self._root,filename)):
                    continue

                suffix = filename[filename.rfind("."):]
                if suffix in imgFormat :
                    imgs.append(filename)

            for filename in sorted(os.listdir(self._root)):
                if osp.isdir(osp.join(self._root,filename)):
                    continue
                suffix = filename[filename.rfind("."):]
                if suffix == ".json":
                    mask_path = osp.join(self._root, filename)

                    with open(mask_path, "r") as f:
                        content = json.load(f)
                    ip = content["imagePath"]
                    if ip in imgs:
                        jsons.append(content["shapes"])
                        json_imgs.append(ip)


            self.imgs = json_imgs
            self._jsons = jsons
        else:
            imgs = []
            for filename in sorted(os.listdir(self._root)):
                if osp.isdir(osp.join(self._root,filename)):
                    continue
                suffix = filename[filename.rfind("."):]
                if suffix in imgFormat:
                    imgs.append(filename)
            self.imgs = imgs

        print("LabelmeDataset Size: ", len(self.imgs))
        # print("Dataset prepared")

    def preprocessLabels(self):
        """将数组转换为 numpy 格式，
        并添加 type 字段表明该区域为 mask 或者 keypoint
        """
        jsons = []
        for cur_label in self._jsons:
            nj = {}
            # for label in cur_label:
                # if 
        pass

    def getMask(self, idx):
        mask_path = osp.join(self.root, "SegmentationClassPNG", self.masks[idx])
        mask = Image.open(mask_path)
        mask = np.array(mask)
        return mask

    def getTarget(self, idx, mask, to_tensor=True):
        """
        Args:
            idx  图片序号

        All values are numpy datatype, Not converted to Tensor
        """
        # instances are encoded as different colors
        obj_ids = np.unique(mask)
        # first id is the background, so remove it
        obj_ids = obj_ids[1:]
        masks = (mask == obj_ids[:, None, None])
        # get bounding box coordinates for each mask
        num_objs = len(obj_ids)
        boxes = []
        labels = []
        iscrowd = []
        for i in range(num_objs):
            pos = np.where(masks[i])
            label = np.max(mask[masks[i]])
            # print(obj_ids, label)
            xmin = np.min(pos[1])
            xmax = np.max(pos[1])
            ymin = np.min(pos[0])
            ymax = np.max(pos[0])
            boxes.append([xmin, ymin, xmax, ymax])
            labels.append(label)
            iscrowd.append(0)
        # print(boxes)

        target = {}

        # target["maskimg"] = mask
        if to_tensor:
            target["image_id"] = torch.tensor([idx])
            boxes = torch.as_tensor(boxes, dtype=torch.float32)
            target["boxes"] = boxes
            # two class currently
            target["labels"] = torch.tensor(labels, dtype=torch.int64)
            target["masks"] = torch.as_tensor(masks, dtype=torch.uint8)
            # target["area"] = torch.tensor(target["area"], dtype=torch.uint8)
            target["area"] = (boxes[:, 3] - boxes[:, 1]) * (boxes[:, 2] - boxes[:, 0])
            target["iscrowd"] = torch.as_tensor(iscrowd, dtype=torch.int64)
        else:
            target["boxes"] = boxes
            target["labels"] = labels
            target["masks"] = masks
            target["image_id"] = [idx]
            target["iscrowd"] = iscrowd

        return target

    def __getitem__(self, idx):
        if not self.isTrain:
            img_path = osp.join(self._root, self.imgs[idx])   
            img = Image.open(img_path)

            img = np.array(img)
            # img = torch.Tensor(img)
            if self.transforms is not None:
                img, _ = self.transforms(img)
            else:
                img = FT.to_tensor(img)
            return img, None

        mask = self._jsons[idx]
        print(mask)
        mask = np.array(mask)
        target = self.getTarget(idx, mask)

        img_path = osp.join(self._root, "JPEGImages", self.imgs[idx])
        img = Image.open(img_path).convert('L')
        img = np.array(img)

        if self.transforms is not None:
            img, target = self.transforms(img, target)
        else:
            img = FT.to_tensor(img)

        return img, target


    def __len__(self):
        return len(self.imgs)

    def __repr__(self):
        return f'<DS {self.name}>'



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""A command which may simplify the creation or deletion a shared git repository on personal mini git server.

"""

__version__ = "0.0.18"

import shutil
import os
import sys
from pathlib import Path
from argparse import ArgumentParser
import json
from bffacilities._frame import Frame
from bffacilities.utils import createLogger, initGetText
from bffacilities._constants import _CONFIG_DIR

tr = initGetText("gitrepo")
logger = createLogger("gitrepo.log", stream=True)

DEFAULT_REPO_DIR = "/src"


class GitRepoFrame(Frame):
    ConfigFile = _CONFIG_DIR / "gitrepo.json"

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.config = {
            "repo_dir": DEFAULT_REPO_DIR,
            "recommended": False,
            "user": "git",
            "group": "git",
        }
    
    def _initConfigFile(self, file):
        if not file.exists():
            with file.open("w") as f:
                json.dump(self.config, f)
            return
    def readConfig(self, file):
        self._initConfigFile(file)
        try:
            with file.open() as f:
                self.config = json.loads(f.read())
        except:
            self._initConfigFile(file)

        if "repo_dir" not in self.config:
            self.config["repo_dir"] = DEFAULT_REPO_DIR

        if "recommended" not in self.config:
            self.config["recommended"] = False

    def initArgs(self):
        self.readConfig(GitRepoFrame.ConfigFile)
        parser = ArgumentParser(prog='bffacility gitrepo', description=tr("""Run this script under SUPER USER privilege. Create or remove git repository
    under direcotry: {}. Currently only works for Linux. Most operations are only tested under Linux.
    Script version : {}.
        """).format(self.config["repo_dir"], __version__))

        parser.add_argument('repo', nargs="?", type=str, help=tr("Specify the repository name, for example: \
            Test.git, if .git suffix is not typed, this script will add it automatically"))

        action = parser.add_mutually_exclusive_group()
        action.add_argument("-i", "--init",
                            help=tr("Default action, Init a new git repository under {}.").format(
                                self.config["repo_dir"]),
                            action="store_true")

        action.add_argument("-d", "--delete",
                            help=tr("Remove a existing repository under {}.").format(
                                self.config["repo_dir"]),
                            action="store_true")

        action.add_argument("-l", "--list",
                            help=tr("List existing repositories under {}.").format(
                                self.config["repo_dir"]),
                            action="store_true")

        parser.add_argument("-s", "--src-dir", help=tr("Specify the src folder for git repository, for example, the default path is `/src`.\
            Do NOT put single-slash `/` as the src folder, it will make unestimated error."))
        parser.add_argument("-u", "--user", help=tr(
            "Specify the git repository owner, for example, the default owner is `git`."))
        parser.add_argument("-g", "--group", help=tr(
            "Specify the git repository owner group, for example, the default owner group is `git`."))
        parser.add_argument("-w", "--windows", help=tr(
            "Force this script running on Windows platform."), action="store_true")
        parser.add_argument("-f", "--force", help=tr(
            "Force this script running even without SUPER USER previlige."), action="store_true")
        parser.add_argument( "-V", "--version", 
            help=tr("Print this script version"), action="version", version=f'gitrepo {__version__}')
        self.parser = parser


    def parseArgs(self, argv):
        args = self.parser.parse_args(argv)
        config = self.config
        config["action"] = "init"
        if args.delete:
            config["action"] = "delete"
        elif args.list:
            config["action"] = "list"

        config["windows"] = args.windows
        config["normal_user"] = args.force

        if config["action"] != "list":
            if not args.repo:
                logger.warning(
                    "Repository name should be specified. Or -l (--list) options should be specified. \n------------------------")
                self.parser.print_help()
                sys.exit(1)
        else:
            config["windows"] = True
            config["normal_user"] = True

        if args.src_dir is not None and args.src_dir != "/":
            config["src"] = args.src_dir

        repo = args.repo or "Test"
        if repo[-4:] != ".git":
            repo = "{}.git".format(repo)

        config["repo"] = repo

    def run(self):
        config = self.config
        repo = Path(config["repo_dir"]) / config["repo"]
        try:
            if config["action"] == "delete":
                deleteRepo(repo)
            elif config["action"] == "init":
                createRepo(repo)
            else:
                listRepo(repo)
        except Exception as e:
            logger.warning(e)

def createRepo(repo: Path):
    if not repo.exists():
        return
    logger.warning("Repository: {} exists, if you want to recreate a new repository, please delete it first.\
        Now change the owner of that repository".format(repo))
    try:
        shutil.chown(repo, config["user"], config["group"])
        logger.info("Repository owner changed.")
    except Exception as e:
        logger.critical(
            "You may need to run this script under super user privilege.")
        logger.critical(e)
        

    logger.info("Creating an repository: {}".format(repo))
    repo.mkdir(parents=True)
    import subprocess
    try:
        subprocess.run(
            ["git", "init", "--bare", "--shared", str(repo.absolute())])
        shutil.chown(repo, config["user"], config["group"])
    except Exception as e:
        logger.critical("Git may not be installed, please install git first, \
            or you are not running this command under super user privilege. You may need to run this script under super user privilege.")
        logger.critical(e)

    logger.info("Repository created.")


def deleteRepo(repo: Path):
    if not repo.exists():
        logger.warning("Repository: {} not exists".format(repo))
        return
    logger.info("Deleteing an existing repository: {}".format(repo))
    try:
        # repo.rmdir()
        shutil.rmtree(str(repo))
        logger.info("Repository deleted.")
    except Exception as e:
        logger.warning(e)


def listRepo(repo: Path):
    repo_dir = repo.parent
    repos = []
    for r in repo_dir.iterdir():
        if r.suffix == ".git" and r.is_dir():
            head = r / "HEAD"
            if head.exists():
                repos.append(r)
    logger.info("Repositories found count: {}".format(len(repos)))
    for re in repos:
        logger.info("Found a repository: {}".format(re.stem))


def is_root():
    if hasattr(os, "getuid"):
        return os.getuid() == 0
    return False


def main(argv = None):
    gp = GitRepoFrame()
    gp.initArgs()
    if argv is None:
        argv = sys.argv[1:]
    gp.parseArgs(argv)
    if not sys.platform.startswith('linux') and not config["windows"]:
        logger.warning("Sorry, this script can only run on linux systems.")
        return

    if not is_root() and not config["normal_user"]:
        logger.warning(
            "Try running this script as normal user.\n---------------")
        parser.print_help()
        return
    gp.run()


def first_taste():
    # for fisrt using this script
    if config["recommended"] is False:
        parser.print_help()
        with GitRepoFrame.ConfigFile.open("w") as f:
            config["recommended"] = True
            json.dump(config, f)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Description: There are two main purpose of this module:
1. provide sqlalchemy as DB to other sub modules so they can define 
their database models even the database instance is not created yet.
2. provide a function called ``createDatabase`` to attach database to 
sqlalchemy, so it can get access to these databases (which would be sqlite
or mysql database).

Author: BriFuture

Modified: 2020/05/13 20:56
解决了 mysql 中 lost connection 的错误
@refrence 
    1. https://towardsdatascience.com/use-flask-and-sqlalchemy-not-flask-sqlalchemy-5a64fafe22a4
    2. https://github.com/pallets/flask-sqlalchemy/
    3. https://docs.sqlalchemy.org/en/13/orm/contextual.html

Modified: 2020/06/19 replace property session with scoped_session_maker
"""


"""
from . import app

app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///%s' % SQLITE_DATABASE_LOC
# app.config['SQLALCHEMY_COMMIT_ON_TEARDOWN'] = True
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True

from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy( app )
"""

__version__ = '0.2.0'
import math

def paginate(self, page=None, per_page=None, to_dict=True):
    """
    分页函数
    :param self:
    :param page:
    :param per_page:
    :return:
    """
    if page is None:
        page = 1

    if per_page is None:
        per_page = 20

    items = self.limit(per_page).offset((page - 1) * per_page).all()

    if not items and page != 1:
        return {'total': 0, 'page': page, 'error': 'no such items'}
        
    if page == 1 and len(items) < per_page:
        total = len(items)
    else:
        total = self.order_by(None).count()
    
    if to_dict:
        ditems = [item.to_dict() for item in items]
    else:
        ditems = items

    return {
        'page': page, 
        'per_page': per_page, 
        'total': total, 
        'items': ditems
    }
    # return Pagination(self, page, per_page, total, items)

import sqlalchemy
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base, as_declarative, declared_attr
# Model = declarative_base()

@as_declarative()
class _Model(object):
    # @declared_attr
    # def query():
    #     return 
    query = None
    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()
    # id = Column(Integer, primary_key=True)
sqlalchemy.Model = _Model
orm.Query.paginate = paginate

# Database Proxy
import time
import logging, sys
import threading
from bffacilities import createLogger

class Database(object):
    """Recommond Usage: 
    ```
    db = Database(dbname)
    with db as sess:
        sess.add(Record())
        sess.commit()
    ```
    But it is compatible with flask_sqlalchemy, 
    but scopefunc must be set when construct database instance, @see create_engine
    """
    Column = sqlalchemy.Column
    Integer = sqlalchemy.Integer
    SmallInteger = sqlalchemy.SmallInteger
    BigInteger = sqlalchemy.BigInteger
    Boolean = sqlalchemy.Boolean
    Enum = sqlalchemy.Enum
    Float = sqlalchemy.Float
    Interval = sqlalchemy.Interval
    Numeric = sqlalchemy.Numeric
    PickleType = sqlalchemy.PickleType
    String = sqlalchemy.String
    Text = sqlalchemy.Text
    DateTime = sqlalchemy.DateTime
    Date = sqlalchemy.Date
    Time = sqlalchemy.Time
    Unicode = sqlalchemy.Unicode
    UnicodeText = sqlalchemy.UnicodeText
    LargeBinary = sqlalchemy.LargeBinary
    # MatchType = sqlalchemy.MatchType
    # SchemaType = sqlalchemy.SchemaType
    ARRAY = sqlalchemy.ARRAY
    BIGINT = sqlalchemy.BIGINT
    BINARY = sqlalchemy.BINARY
    BLOB = sqlalchemy.BLOB
    relationship = sqlalchemy.orm.relationship
    ForeignKey = sqlalchemy.ForeignKey
    Table = sqlalchemy.Table

    Model = _Model
    _SessionFactory = orm.sessionmaker()

    def __init__(self, dbname, logger=None, delay_engine_create=False, **kwargs):
        """db used to replace flask_sqlalchemy to provide more convinient ways to manage app
        @pool_recycle 默认的刷新时间为 10 分钟
        @Session use Session instead session property to get scoped session, 
            but scopefunc must be set before Session could be used
        @see createEngine

        kwargs scopefunc
        """
        if logger is None:
            logger = createLogger("db", savefile=False, stream=True, level=logging.DEBUG)
        self.logger = logger

        if 'pool_recycle' not in kwargs:
            kwargs['pool_recycle'] = 600
        # kwargs['pool_recycle'] = 5 # TEST
        self._refresh = kwargs['pool_recycle'] - 1
        
        self.create_all = _Model.metadata.create_all
        self.drop_all = _Model.metadata.drop_all
        if not delay_engine_create:
            self.create_engine(dbname, **kwargs)

    def create_engine(self, dbname, **kwargs):
        """::kwargs:: 
        scopefunc
            if flask is used, `scopefunc=_app_ctx_stack.__ident_func__` could be passed in kwargs
        """
        scopefunc = kwargs.pop('scopefunc', None)
        engine = sqlalchemy.create_engine(dbname, **kwargs)
        _Model.metadata.bind = engine
        
        Database._SessionFactory.configure(bind=engine)
        # _Session is the scoped session maker
        self.Session = orm.scoped_session(Database._SessionFactory, scopefunc=scopefunc)
        # used for compatibility with flask_sqlalchemy, session is scoped
        # self.session = self.Session
        _Model.query = self.Session.query_property()

        self.engine = engine
        self._sessions = {}
        self._sessionIds = []
        # self.singleSession = dbname.startswith("sqlite")
        self.singleSession = False

    @property
    def session(self):
        """compat with flask_sqlalchemy, not recommand 

        Try use 
        ```
        with database as  sess:
            sess.insert...
        ```
        or
        ```
        sess = database.Session()
        
        ...

        sess.close()
        ```

        """
        id = 0 if self.singleSession else threading.get_ident()
        # self.logger.debug(f"GetSession: {id} {len(self._sessions)}")
        removeneed = []
        for kid, sess in self._sessions.items():
            if not self.valid_session(sess):
                removeneed.append((kid, sess))
                self.logger.debug(f"Found Invalid Session: {kid} {sess}")

        for rn in removeneed:
            self._sessions.pop(rn[0])
            try:
                rn[1].close()
            except: 
                self.logger.debug(f"Close Session Error: {rn[0]} {rn[1]}")

        sess = self._sessions.get(id, None)
        if sess is not None:
            return sess
        count = len(self._sessions)
        if count >= 1000:
            sids = self._sessionIds[:count]
            self._sessionIds = self._sessionIds[count:]
            for sid in sids:
                sess = self._sessions.pop(sid)
                sess.close()
            # return
        sess = self.Session()
        self.logger.debug(f"Create Session For: {id}")
        self._sessions[id] = sess
        self._sessionIds.append(id)
        return sess
    
    @staticmethod
    def valid_session(sess):
        try:
            # Try to get the underlying session connection, If you can get it, it's up
            connection = sess.connection()
            # connection.close()
        except:
            return False
        return True

    def close(self):
        self.Session.remove()

    def __enter__(self):
        self._scopesession = self.Session()
        return self._scopesession

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._scopesession.close()
        if exc_type:
            self.logger.warning(f"Type: {exc_type}, value: {exc_val}, {exc_tb}")


# Database Proxy
def createDatabase(dbname, **kwargs):
    """create instance of sqlalchemy Database
    """
    db = Database(dbname, **kwargs)
    logger = logging.getLogger('db')
    logger.warning("[DB] Use Class Database directly")
    return db


import json
import os
import os.path as osp
from tqdm import tqdm
from PIL import Image
import numpy as np
# from labelme.label_file import LabelFile, PY2, QT4, utils, io
import shutil as sh
import glob
import imgviz
import labelme

__version__ = "0.1.1"


class Labelme2Vocor():

    def __init__(self, input_dir, output_dir, noviz=False, debug=False, **kwargs):
        if osp.exists(output_dir):
            print("****Output directory already exists: ", output_dir, " They may need to ne deleted...")
            # sh.rmtree(output_dir)
            # return
        os.makedirs(output_dir, exist_ok=True)
        os.makedirs(osp.join(output_dir, "JPEGImages"), exist_ok=True)
        os.makedirs(osp.join(output_dir, "SegmentationClass"), exist_ok=True)
        os.makedirs(osp.join(output_dir, "SegmentationClassPNG"), exist_ok=True)

        if not noviz:
            os.makedirs(osp.join(output_dir, "SegmentationClassVisualization"), exist_ok=True)
        print("Creating dataset:", output_dir)

        self.input_dir = input_dir
        self.output_dir = output_dir
        self.noviz = noviz
        self.debug = debug

    def setClasses(self, classes):
        """不含 background
        """
        self.class_names = classes
        self.class_name_to_id = {}
        for i, cn in enumerate(classes):
            self.class_name_to_id[cn] = i + 1
        self.class_name_to_id["_background_"] = 0
        self._write_out()

    def getClassesFromFile(self, classfile):
        """
        根据文件获取 class

        ```
        __ignore__
        _background_
        class1
        class2
        ```
        """
        ## get class names
        contents = []
        with open(classfile, "r") as f:
            for line in f:
                line = line.strip()
                if len(line) == 0 or line.startswith("#"):
                    continue
                contents.append(line)

        class_names = []
        class_name_to_id = {}  # name: 1
        for i, line in enumerate(contents):
            class_id = i - 1  # starts with -1
            class_name = line.strip()
            if len(class_name) == 0:
                continue
            class_name_to_id[class_name] = class_id
            if class_id == -1:
                assert class_name == "__ignore__"
                continue
            elif class_id == 0:
                assert class_name == "_background_"
            class_names.append(class_name)

        class_names = tuple(class_names)

        self.class_names = class_names
        self.class_name_to_id = class_name_to_id
        self._write_out()

    def _write_out(self):
        out_class_names_file = osp.join(self.output_dir, "class_names.txt")

        with open(out_class_names_file, "w") as f:
            f.writelines("\n".join(self.class_names))
        print("Class_names:", self.class_names)
        print("Saved class_names:", out_class_names_file)

    def output(self, ordered_keys=None):
        """
        Arguments:
            ordered_keys(list(str))
        """
        print("Generating dataset from:", self.input_dir, " ====> ", self.output_dir)
        for filename in tqdm(glob.glob(osp.join(self.input_dir, "*.json"))):

            label_file = labelme.LabelFile(filename=filename)
            base = osp.splitext(osp.basename(filename))[0]

            out_img_file = osp.join(self.output_dir, "JPEGImages", base + ".jpg")

            with open(out_img_file, "wb") as f:
                f.write(label_file.imageData)

            if ordered_keys is not None:
                newshapes = []
                for ok in ordered_keys:
                    for shape in label_file.shapes:
                        if shape["label"] == ok:
                            newshapes.append(shape)
                label_file.shapes=newshapes

            img = labelme.utils.img_data_to_arr(label_file.imageData)

            lbl, _ = labelme.utils.shapes_to_label(
                img_shape=img.shape,
                shapes=label_file.shapes,
                label_name_to_value=self.class_name_to_id,
            )

            out_png_file = osp.join(self.output_dir, "SegmentationClassPNG", base + ".png")
            labelme.utils.lblsave(out_png_file, lbl)

            # out_lbl_file = osp.join(self.output_dir, "SegmentationClass", base + ".npy")
            # np.save(out_lbl_file, lbl)
                
            if not self.noviz:
                out_viz_file = osp.join(self.output_dir, "SegmentationClassVisualization", base + ".jpg",)

                if img.shape[0] == 1: # gray img
                    img = imgviz.rgb2gray(img)
                h, w = img.shape[:2]
                
                viz = imgviz.label2rgb(
                    label=lbl,
                    # img=imgviz.rgb2gray(img),
                    img=img,
                    font_size=max(10, w // 10),
                    label_names=self.class_names,
                    loc="rb",
                )
                imgviz.io.imsave(out_viz_file, viz)

            
            if self.debug:
                print(label_file.shapes)
                break

def main(args = None):
    from argparse import ArgumentParser
    parser = ArgumentParser("labelme2voc")
    parser.add_argument("--input", type=str, required=True, help="Input Directory contains *.json annotation")
    parser.add_argument("--output", type=str, required=True, help="Output Directory, should be different from input")
    parser.add_argument("--classfile", type=str, default="class_names.txt",  help="Specify file contains class names" )
    parser.add_argument("--orderkeys", type=str, default=None, help="Orderedkyes for classes in label file" )

    args = vars(parser.parse_args(args))
    voc = Labelme2Vocor(args["input"], args["output"])
    voc.getClassesFromFile(args["classfile"])
    voc.output(args["orderkeys"])

if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
from torch.autograd import Variable

from collections import OrderedDict
import numpy as np


def summary(model, input_size, batch_size=-1, device="cuda"):

    def register_hook(module):

        def hook(module, input, output):
            class_name = str(module.__class__).split(".")[-1].split("'")[0]
            module_idx = len(summary)

            m_key = "%s-%i" % (class_name, module_idx + 1)
            summary[m_key] = OrderedDict()

            try:
                summary[m_key]["input_shape"] = list(input[0].size())
            except:
                summary[m_key]["input_shape"] = list((0, 0, 0))
            summary[m_key]["input_shape"][0] = batch_size
                
            if isinstance(output, (list, tuple)):
                try:
                    summary[m_key]["output_shape"] = [
                        [-1] + list(o.size())[1:] for o in output
                    ]
                except:
                    summary[m_key]["output_shape"] = []
            elif isinstance(output, (OrderedDict,)):
                summary[m_key]["output_shape"] = [
                    [-1] + list(o.size())[1:] for _, o in output.items()
                ]
            else:
                summary[m_key]["output_shape"] = list(output.size())
                summary[m_key]["output_shape"][0] = batch_size

            params = 0
            if hasattr(module, "weight") and hasattr(module.weight, "size"):
                params += torch.prod(torch.LongTensor(list(module.weight.size())))
                summary[m_key]["trainable"] = module.weight.requires_grad
            if hasattr(module, "bias") and hasattr(module.bias, "size"):
                params += torch.prod(torch.LongTensor(list(module.bias.size())))
            summary[m_key]["nb_params"] = params

        if (
            not isinstance(module, nn.Sequential)
            and not isinstance(module, nn.ModuleList)
            and not (module == model)
        ):
            hooks.append(module.register_forward_hook(hook))

    device = device.lower()
    assert device in [
        "cuda",
        "cpu",
    ], "Input device is not valid, please specify 'cuda' or 'cpu'"

    if device == "cuda" and torch.cuda.is_available():
        dtype = torch.cuda.FloatTensor
    else:
        dtype = torch.FloatTensor

    # multiple inputs to the network
    if isinstance(input_size, tuple):
        input_size = [input_size]

    # batch_size of 2 for batchnorm
    x = [torch.rand(2, *in_size).type(dtype) for in_size in input_size]
    # print(type(x[0]))

    # create properties
    summary = OrderedDict()
    hooks = []

    # register hook
    model.apply(register_hook)

    # make a forward pass
    # print(x.shape)
    model(*x)

    # remove these hooks
    for h in hooks:
        h.remove()

    print("----------------------------------------------------------------")
    line_new = "{:>20}  {:>25} {:>15}".format("Layer (type)", "Output Shape", "Param #")
    print(line_new)
    print("================================================================")
    total_params = 0
    total_output = 0
    trainable_params = 0
    for layer in summary:
        # input_shape, output_shape, trainable, nb_params
        line_new = "{:>20}  {:>25} {:>15}".format(
            layer,
            str(summary[layer]["output_shape"]),
            "{0:,}".format(summary[layer]["nb_params"]),
        )
        total_params += summary[layer]["nb_params"]
        try:
            total_output += np.prod(summary[layer]["output_shape"])
            # print(summary[layer]["output_shape"])
        except: pass
            
        if "trainable" in summary[layer]:
            if summary[layer]["trainable"] == True:
                trainable_params += summary[layer]["nb_params"]
        print(line_new)

    # assume 4 bytes/number (float on cuda).
    total_input_size = abs(np.prod(input_size) * batch_size * 4. / (1024 ** 2.))
    total_output_size = abs(2. * total_output * 4. / (1024 ** 2.))  # x2 for gradients
    total_params_size = abs(total_params.numpy() * 4. / (1024 ** 2.))
    total_size = total_params_size + total_output_size + total_input_size

    print("================================================================")
    print("Total params: {0:,}".format(total_params))
    print("Trainable params: {0:,}".format(trainable_params))
    print("Non-trainable params: {0:,}".format(total_params - trainable_params))
    print("----------------------------------------------------------------")
    print("Input size (MB): %0.2f" % total_input_size)
    print("Forward/backward pass size (MB): %0.2f" % total_output_size)
    print("Params size (MB): %0.2f" % total_params_size)
    print("Estimated Total Size (MB): %0.2f" % total_size)
    print("----------------------------------------------------------------")
    # return summary



# -*- coding: utf-8 -*-
"""
    Server-side Sessions and SessionInterfaces.

    :copyright: (c) 2014 by Shipeng Feng.
    :license: BSD, see LICENSE for more details.

    This file is copied from flask_session and modified by brifuture for werkzeug has upgraded to v1.0,
    memcache is not avialable from werkzueg but from cachelib
"""

import sys
import time
from datetime import datetime
from uuid import uuid4
try:
    import cPickle as pickle
except ImportError:
    import pickle

from flask.sessions import SessionInterface as FlaskSessionInterface
from flask.sessions import SessionMixin
from werkzeug.datastructures import CallbackDict
from itsdangerous import Signer, BadSignature, want_bytes


PY2 = sys.version_info[0] == 2
if not PY2:
    text_type = str
else:
    text_type = unicode


def total_seconds(td):
    return td.days * 60 * 60 * 24 + td.seconds


class ServerSideSession(CallbackDict, SessionMixin):
    """Baseclass for server-side based sessions."""

    def __init__(self, initial=None, sid=None, permanent=None):
        def on_update(self):
            self.modified = True
        CallbackDict.__init__(self, initial, on_update)
        self.sid = sid
        if permanent:
            self.permanent = permanent
        self.modified = False


class RedisSession(ServerSideSession):
    pass


class MemcachedSession(ServerSideSession):
    pass


class FileSystemSession(ServerSideSession):
    pass


class MongoDBSession(ServerSideSession):
    pass


class SqlAlchemySession(ServerSideSession):
    pass


class SessionInterface(FlaskSessionInterface):

    def _generate_sid(self):
        return str(uuid4())

    def _get_signer(self, app):
        if not app.secret_key:
            return None
        return Signer(app.secret_key, salt='flask-session',
                      key_derivation='hmac')


class NullSessionInterface(SessionInterface):
    """Used to open a :class:`flask.sessions.NullSession` instance.
    """

    def open_session(self, app, request):
        return None


class RedisSessionInterface(SessionInterface):
    """Uses the Redis key-value store as a session backend.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.

    :param redis: A ``redis.Redis`` instance.
    :param key_prefix: A prefix that is added to all Redis store keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    """

    serializer = pickle
    session_class = RedisSession

    def __init__(self, redis, key_prefix, use_signer=False, permanent=True):
        if redis is None:
            from redis import Redis
            redis = Redis()
        self.redis = redis
        self.key_prefix = key_prefix
        self.use_signer = use_signer
        self.permanent = permanent

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        if not sid:
            sid = self._generate_sid()
            return self.session_class(sid=sid, permanent=self.permanent)
        if self.use_signer:
            signer = self._get_signer(app)
            if signer is None:
                return None
            try:
                sid_as_bytes = signer.unsign(sid)
                sid = sid_as_bytes.decode()
            except BadSignature:
                sid = self._generate_sid()
                return self.session_class(sid=sid, permanent=self.permanent)

        if not PY2 and not isinstance(sid, text_type):
            sid = sid.decode('utf-8', 'strict')
        val = self.redis.get(self.key_prefix + sid)
        if val is not None:
            try:
                data = self.serializer.loads(val)
                return self.session_class(data, sid=sid)
            except:
                return self.session_class(sid=sid, permanent=self.permanent)
        return self.session_class(sid=sid, permanent=self.permanent)

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        path = self.get_cookie_path(app)
        if not session:
            if session.modified:
                self.redis.delete(self.key_prefix + session.sid)
                response.delete_cookie(app.session_cookie_name,
                                       domain=domain, path=path)
            return

        # Modification case.  There are upsides and downsides to
        # emitting a set-cookie header each request.  The behavior
        # is controlled by the :meth:`should_set_cookie` method
        # which performs a quick check to figure out if the cookie
        # should be set or not.  This is controlled by the
        # SESSION_REFRESH_EACH_REQUEST config flag as well as
        # the permanent flag on the session itself.
        # if not self.should_set_cookie(app, session):
        #    return

        httponly = self.get_cookie_httponly(app)
        secure = self.get_cookie_secure(app)
        expires = self.get_expiration_time(app, session)
        val = self.serializer.dumps(dict(session))
        self.redis.setex(name=self.key_prefix + session.sid, value=val,
                         time=total_seconds(app.permanent_session_lifetime))
        if self.use_signer:
            session_id = self._get_signer(app).sign(want_bytes(session.sid))
        else:
            session_id = session.sid
        response.set_cookie(app.session_cookie_name, session_id,
                            expires=expires, httponly=httponly,
                            domain=domain, path=path, secure=secure)


class MemcachedSessionInterface(SessionInterface):
    """A Session interface that uses memcached as backend.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.

    :param client: A ``memcache.Client`` instance.
    :param key_prefix: A prefix that is added to all Memcached store keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    """

    serializer = pickle
    session_class = MemcachedSession

    def __init__(self, client, key_prefix, use_signer=False, permanent=True):
        if client is None:
            client = self._get_preferred_memcache_client()
            if client is None:
                raise RuntimeError('no memcache module found')
        self.client = client
        self.key_prefix = key_prefix
        self.use_signer = use_signer
        self.permanent = permanent

    def _get_preferred_memcache_client(self):
        servers = ['127.0.0.1:11211']
        try:
            import pylibmc
        except ImportError:
            pass
        else:
            return pylibmc.Client(servers)

        try:
            import memcache
        except ImportError:
            pass
        else:
            return memcache.Client(servers)

    def _get_memcache_timeout(self, timeout):
        """
        Memcached deals with long (> 30 days) timeouts in a special
        way. Call this function to obtain a safe value for your timeout.
        """
        if timeout > 2592000:  # 60*60*24*30, 30 days
            # See http://code.google.com/p/memcached/wiki/FAQ
            # "You can set expire times up to 30 days in the future. After that
            # memcached interprets it as a date, and will expire the item after
            # said date. This is a simple (but obscure) mechanic."
            #
            # This means that we have to switch to absolute timestamps.
            timeout += int(time.time())
        return timeout

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        if not sid:
            sid = self._generate_sid()
            return self.session_class(sid=sid, permanent=self.permanent)
        if self.use_signer:
            signer = self._get_signer(app)
            if signer is None:
                return None
            try:
                sid_as_bytes = signer.unsign(sid)
                sid = sid_as_bytes.decode()
            except BadSignature:
                sid = self._generate_sid()
                return self.session_class(sid=sid, permanent=self.permanent)

        full_session_key = self.key_prefix + sid
        if PY2 and isinstance(full_session_key, unicode):
            full_session_key = full_session_key.encode('utf-8')
        val = self.client.get(full_session_key)
        if val is not None:
            try:
                if not PY2:
                    val = want_bytes(val)
                data = self.serializer.loads(val)
                return self.session_class(data, sid=sid)
            except:
                return self.session_class(sid=sid, permanent=self.permanent)
        return self.session_class(sid=sid, permanent=self.permanent)

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        path = self.get_cookie_path(app)
        full_session_key = self.key_prefix + session.sid
        if PY2 and isinstance(full_session_key, unicode):
            full_session_key = full_session_key.encode('utf-8')
        if not session:
            if session.modified:
                self.client.delete(full_session_key)
                response.delete_cookie(app.session_cookie_name,
                                       domain=domain, path=path)
            return

        httponly = self.get_cookie_httponly(app)
        secure = self.get_cookie_secure(app)
        expires = self.get_expiration_time(app, session)
        if not PY2:
            val = self.serializer.dumps(dict(session), 0)
        else:
            val = self.serializer.dumps(dict(session))
        self.client.set(full_session_key, val, self._get_memcache_timeout(
                        total_seconds(app.permanent_session_lifetime)))
        if self.use_signer:
            session_id = self._get_signer(app).sign(want_bytes(session.sid))
        else:
            session_id = session.sid
        response.set_cookie(app.session_cookie_name, session_id,
                            expires=expires, httponly=httponly,
                            domain=domain, path=path, secure=secure)


class FileSystemSessionInterface(SessionInterface):
    """Uses the :class:`werkzeug.contrib.cache.FileSystemCache` as a session
    backend.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.

    :param cache_dir: the directory where session files are stored.
    :param threshold: the maximum number of items the session stores before it
                      starts deleting some.
    :param mode: the file mode wanted for the session files, default 0600
    :param key_prefix: A prefix that is added to FileSystemCache store keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    """

    session_class = FileSystemSession

    def __init__(self, cache_dir, threshold, mode, key_prefix,
                 use_signer=False, permanent=True):
        from cachelib import FileSystemCache
        self.cache = FileSystemCache(cache_dir, threshold=threshold, mode=mode)
        self.key_prefix = key_prefix
        self.use_signer = use_signer
        self.permanent = permanent

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        if not sid:
            sid = self._generate_sid()
            return self.session_class(sid=sid, permanent=self.permanent)
        if self.use_signer:
            signer = self._get_signer(app)
            if signer is None:
                return None
            try:
                sid_as_bytes = signer.unsign(sid)
                sid = sid_as_bytes.decode()
            except BadSignature:
                sid = self._generate_sid()
                return self.session_class(sid=sid, permanent=self.permanent)

        data = self.cache.get(self.key_prefix + sid)
        if data is not None:
            return self.session_class(data, sid=sid)
        return self.session_class(sid=sid, permanent=self.permanent)

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        path = self.get_cookie_path(app)
        if not session:
            if session.modified:
                self.cache.delete(self.key_prefix + session.sid)
                response.delete_cookie(app.session_cookie_name,
                                       domain=domain, path=path)
            return

        httponly = self.get_cookie_httponly(app)
        secure = self.get_cookie_secure(app)
        expires = self.get_expiration_time(app, session)
        data = dict(session)
        self.cache.set(self.key_prefix + session.sid, data,
                       total_seconds(app.permanent_session_lifetime))
        if self.use_signer:
            session_id = self._get_signer(app).sign(want_bytes(session.sid))
        else:
            session_id = session.sid
        response.set_cookie(app.session_cookie_name, session_id,
                            expires=expires, httponly=httponly,
                            domain=domain, path=path, secure=secure)


class MongoDBSessionInterface(SessionInterface):
    """A Session interface that uses mongodb as backend.

    .. versionadded:: 0.2
        The `use_signer` parameter was added.

    :param client: A ``pymongo.MongoClient`` instance.
    :param db: The database you want to use.
    :param collection: The collection you want to use.
    :param key_prefix: A prefix that is added to all MongoDB store keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    """

    serializer = pickle
    session_class = MongoDBSession

    def __init__(self, client, db, collection, key_prefix, use_signer=False,
                 permanent=True):
        if client is None:
            from pymongo import MongoClient
            client = MongoClient()
        self.client = client
        self.store = client[db][collection]
        self.key_prefix = key_prefix
        self.use_signer = use_signer
        self.permanent = permanent

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        if not sid:
            sid = self._generate_sid()
            return self.session_class(sid=sid, permanent=self.permanent)
        if self.use_signer:
            signer = self._get_signer(app)
            if signer is None:
                return None
            try:
                sid_as_bytes = signer.unsign(sid)
                sid = sid_as_bytes.decode()
            except BadSignature:
                sid = self._generate_sid()
                return self.session_class(sid=sid, permanent=self.permanent)

        store_id = self.key_prefix + sid
        document = self.store.find_one({'id': store_id})
        if document and document.get('expiration') <= datetime.utcnow():
            # Delete expired session
            self.store.remove({'id': store_id})
            document = None
        if document is not None:
            try:
                val = document['val']
                data = self.serializer.loads(want_bytes(val))
                return self.session_class(data, sid=sid)
            except:
                return self.session_class(sid=sid, permanent=self.permanent)
        return self.session_class(sid=sid, permanent=self.permanent)

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        path = self.get_cookie_path(app)
        store_id = self.key_prefix + session.sid
        if not session:
            if session.modified:
                self.store.remove({'id': store_id})
                response.delete_cookie(app.session_cookie_name,
                                       domain=domain, path=path)
            return

        httponly = self.get_cookie_httponly(app)
        secure = self.get_cookie_secure(app)
        expires = self.get_expiration_time(app, session)
        val = self.serializer.dumps(dict(session))
        self.store.update({'id': store_id},
                          {'id': store_id,
                           'val': val,
                           'expiration': expires}, True)
        if self.use_signer:
            session_id = self._get_signer(app).sign(want_bytes(session.sid))
        else:
            session_id = session.sid
        response.set_cookie(app.session_cookie_name, session_id,
                            expires=expires, httponly=httponly,
                            domain=domain, path=path, secure=secure)


class SqlAlchemySessionInterface(SessionInterface):
    """Uses the Flask-SQLAlchemy from a flask app as a session backend.

    .. versionadded:: 0.2

    :param app: A Flask app instance.
    :param db: A Flask-SQLAlchemy instance.
    :param table: The table name you want to use.
    :param key_prefix: A prefix that is added to all store keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    """

    serializer = pickle
    session_class = SqlAlchemySession

    def __init__(self, app, db, table, key_prefix, use_signer=False,
                 permanent=True):
        if db is None:
            from flask_sqlalchemy import SQLAlchemy
            db = SQLAlchemy(app)
        self.db = db
        self.key_prefix = key_prefix
        self.use_signer = use_signer
        self.permanent = permanent

        class Session(self.db.Model):
            __tablename__ = table

            id = self.db.Column(self.db.Integer, primary_key=True)
            session_id = self.db.Column(self.db.String(255), unique=True)
            data = self.db.Column(self.db.LargeBinary)
            expiry = self.db.Column(self.db.DateTime)

            def __init__(self, session_id, data, expiry):
                self.session_id = session_id
                self.data = data
                self.expiry = expiry

            def __repr__(self):
                return '<Session data %s>' % self.data

        # self.db.create_all()
        self.sql_session_model = Session

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        if not sid:
            sid = self._generate_sid()
            return self.session_class(sid=sid, permanent=self.permanent)
        if self.use_signer:
            signer = self._get_signer(app)
            if signer is None:
                return None
            try:
                sid_as_bytes = signer.unsign(sid)
                sid = sid_as_bytes.decode()
            except BadSignature:
                sid = self._generate_sid()
                return self.session_class(sid=sid, permanent=self.permanent)

        store_id = self.key_prefix + sid
        saved_session = self.sql_session_model.query.filter_by(session_id=store_id).first()
        if saved_session and saved_session.expiry <= datetime.utcnow():
            # Delete expired session
            self.db.session.delete(saved_session)
            self.db.session.commit()
            saved_session = None
        if saved_session:
            try:
                val = saved_session.data
                data = self.serializer.loads(want_bytes(val))
                return self.session_class(data, sid=sid)
            except:
                return self.session_class(sid=sid, permanent=self.permanent)
        return self.session_class(sid=sid, permanent=self.permanent)

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        path = self.get_cookie_path(app)
        store_id = self.key_prefix + session.sid
        saved_session = self.sql_session_model.query.filter_by(
            session_id=store_id).first()
        if not session:
            if session.modified:
                if saved_session:
                    self.db.session.delete(saved_session)
                    self.db.session.commit()
                response.delete_cookie(app.session_cookie_name,
                                       domain=domain, path=path)
            return

        httponly = self.get_cookie_httponly(app)
        secure = self.get_cookie_secure(app)
        expires = self.get_expiration_time(app, session)
        val = self.serializer.dumps(dict(session))
        if saved_session:
            saved_session.data = val
            saved_session.expiry = expires
            self.db.session.commit()
        else:
            new_session = self.sql_session_model(store_id, val, expires)
            self.db.session.add(new_session)
            self.db.session.commit()
        if self.use_signer:
            session_id = self._get_signer(app).sign(want_bytes(session.sid))
        else:
            session_id = session.sid
        response.set_cookie(app.session_cookie_name, session_id,
                            expires=expires, httponly=httponly,
                            domain=domain, path=path, secure=secure)


import os

# from .sessions import NullSessionInterface, RedisSessionInterface, \
#     MemcachedSessionInterface, FileSystemSessionInterface, \
#     MongoDBSessionInterface, SqlAlchemySessionInterface


class Session(object):
    """This class is used to add Server-side Session to one or more Flask
    applications.

    There are two usage modes.  One is initialize the instance with a very
    specific Flask application::

        app = Flask(__name__)
        Session(app)

    The second possibility is to create the object once and configure the
    application later::

        sess = Session()

        def create_app():
            app = Flask(__name__)
            sess.init_app(app)
            return app

    By default Flask-Session will use :class:`NullSessionInterface`, you
    really should configurate your app to use a different SessionInterface.

    .. note::

        You can not use ``Session`` instance directly, what ``Session`` does
        is just change the :attr:`~flask.Flask.session_interface` attribute on
        your Flask applications.
    """

    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """This is used to set up session for your app object.

        :param app: the Flask app object with proper configuration.
        """
        app.session_interface = self._get_interface(app)

    def _get_interface(self, app):
        config = app.config.copy()
        config.setdefault('SESSION_TYPE', 'null')
        config.setdefault('SESSION_PERMANENT', True)
        config.setdefault('SESSION_USE_SIGNER', False)
        config.setdefault('SESSION_KEY_PREFIX', 'session:')
        config.setdefault('SESSION_REDIS', None)
        config.setdefault('SESSION_MEMCACHED', None)
        config.setdefault('SESSION_FILE_DIR',
                          os.path.join(os.getcwd(), 'flask_session'))
        config.setdefault('SESSION_FILE_THRESHOLD', 500)
        config.setdefault('SESSION_FILE_MODE', 384)
        config.setdefault('SESSION_MONGODB', None)
        config.setdefault('SESSION_MONGODB_DB', 'flask_session')
        config.setdefault('SESSION_MONGODB_COLLECT', 'sessions')
        config.setdefault('SESSION_SQLALCHEMY', None)
        config.setdefault('SESSION_SQLALCHEMY_TABLE', 'sessions')

        if config['SESSION_TYPE'] == 'redis':
            session_interface = RedisSessionInterface(
                config['SESSION_REDIS'], config['SESSION_KEY_PREFIX'],
                config['SESSION_USE_SIGNER'], config['SESSION_PERMANENT'])
        elif config['SESSION_TYPE'] == 'memcached':
            session_interface = MemcachedSessionInterface(
                config['SESSION_MEMCACHED'], config['SESSION_KEY_PREFIX'],
                config['SESSION_USE_SIGNER'], config['SESSION_PERMANENT'])
        elif config['SESSION_TYPE'] == 'filesystem':
            session_interface = FileSystemSessionInterface(
                config['SESSION_FILE_DIR'], config['SESSION_FILE_THRESHOLD'],
                config['SESSION_FILE_MODE'], config['SESSION_KEY_PREFIX'],
                config['SESSION_USE_SIGNER'], config['SESSION_PERMANENT'])
        elif config['SESSION_TYPE'] == 'mongodb':
            session_interface = MongoDBSessionInterface(
                config['SESSION_MONGODB'], config['SESSION_MONGODB_DB'],
                config['SESSION_MONGODB_COLLECT'],
                config['SESSION_KEY_PREFIX'], config['SESSION_USE_SIGNER'],
                config['SESSION_PERMANENT'])
        elif config['SESSION_TYPE'] == 'sqlalchemy':
            session_interface = SqlAlchemySessionInterface(
                app, config['SESSION_SQLALCHEMY'],
                config['SESSION_SQLALCHEMY_TABLE'],
                config['SESSION_KEY_PREFIX'], config['SESSION_USE_SIGNER'],
                config['SESSION_PERMANENT'])
        else:
            session_interface = NullSessionInterface()

        return session_interface



# 
__version__ = '0.0.5'

import socket
import struct
import argparse
from bffacilities._frame import Frame, common_entry
from bffacilities.utils import createLogger, initGetText
import os.path as osp
from pathlib import Path
import json
_BASE_DIR = Path.home() / ".bfp" # Basic Path

tr = initGetText("wakeonlan")
logger = createLogger("bff", stream=True, savefile=False)

class LanWaker(Frame):
    DefaultLoc = _BASE_DIR / "wakeonlan.txt"
    def __init__(self):
        super().__init__()
    
    def readConfig(self, file):
        pass

    def initArgs(self):
        """ @return ArgumentParser """
        parser = argparse.ArgumentParser(prog="wakeOnLan", description=tr("Used for wake computer on lan."))
        parser.add_argument("-a", "--address", help=tr("specify broadcast ip (v4) address default is 255.255.255.255"), default="255.255.255.255")
        parser.add_argument("-p", "--port", type=int, help=tr("specify port, default is 7"), default=7)
        parser.add_argument("-m", "--mac", type=str, help=tr("specify mac address"), default="FF:FF:FF:FF:FF:FF")
        parser.add_argument("-n", "--name", type=str, help=tr("specify the name of the machine, if specified, it will be recorded."))
        parser.add_argument("-l", "--list", action="store_true", help=tr("List recorded machine."))
        parser.add_argument("-r", "--repeat", type=str, help=tr("Use recorded machine's info."))
        parser.add_argument("-c", "--clear", action="store_true", help=tr("Clear recorded machine's info."))
        parser.add_argument("-V", "--version", help=tr("Print this script version"), 
            action="version", version=f'wakeOnLan {__version__}')

        self.parser = parser

    # def _read(self, )
    def parseArgs(self, argv):
        args = vars(self.parser.parse_args(argv))
        listIt = args["list"]
        repeat = args["repeat"]
        if listIt:
            try:
                with open(LanWaker.DefaultLoc, "r", encoding="utf-8") as f:
                    for l in f:
                        print(l)
            except:
                print("[List] Error while reading local records.")
            return
        if args["clear"]:
            try:
                with open(LanWaker.DefaultLoc, "w", encoding="utf-8") as f:
                    f.write("")
            except:
                pass
            return
        if repeat:
            if self.findName(repeat):
                self.wake(args["address"], args["port"], mac = args["mac"])
            return

        mac = args["mac"]
        mac = mac.replace(":", "")
        mac = mac.replace("-", "")
        mac = mac.upper()
        self.wake(args["address"], args["port"], mac)
        name = args["name"]
        if name is not None and len(name) > 0 and not self.findName(name):
            with open(LanWaker.DefaultLoc, "a", encoding="utf-8") as f:
                content = json.dumps({
                    "mac": mac,
                    "address": args["address"],
                    "port": args["port"],
                    "name": name
                })
                f.write(f"{content}\n")

    def findName(self, name):
        if not osp.exists(LanWaker.DefaultLoc):
            return False
        try:
            with open(LanWaker.DefaultLoc, "r", encoding="utf-8") as f:
                for l in f:
                    l = l.strip()
                    if l.startswith("#") or len(l) == 0:
                        continue
                    args = json.loads(l)
                    if args['name'] == name:
                        return True
                        # break
        except Exception as e:
            print("[FN] Error while reading local records: ", e)
        
        return False
        
    def wake(self, addr, port, mac):
        mac_data = []
        try:
            for i in range(0, 12, 2):
                mac_data.append(int(mac[i:i+2], 16))
        except:
            logger.error(tr("Wrong Mac Address"))
            self.parser.print_help()
        packet = struct.pack("!BBBBBB", 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF)
        packet_mac = struct.pack("!BBBBBB", *mac_data)
        for i in range(0, 16):
            packet += packet_mac
        # print("len: ", len(packet), "data: ", packet)
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        try:
            sock.sendto(packet, (addr, port))
            logger.info(f"{tr('Send wake packet to:')} {addr}:{port}, {mac}")
        finally:
            sock.close()
            
main = common_entry(LanWaker)
